(* Desugaring passes.

  Ordering:
   - LetUnrecursion - transforms recursive lets into let + fix
   - PolyLetExpansion - expand polymorphic let bindings (they bind values !)
   - ANFTransform - match and construct on variables only
   - PatternSimplification - transform deep patterns into shallow patterns

 *)

open Util
open Ast
open! Ast.Explicit
module X = Ast.Explicit

module LetUnrecursion = struct
  (* Transform a multi-binding let rec into a multi-binding let
     binding many multifixes *)
  type mark += Unrecursed

  let extract_lambda (a,_,_) =
    let extr' = function
      | Abs(x,x_ty,((_,e_ty,_) as e)) -> (x,x_ty,e,e_ty)
      | _ -> failwith "No lambda in recursive binding"
    in
    extr' a

  let unrecurse =
    let rec un (e,ty,loc) = (un' loc e, ty, loc)
    and un' loc = function
      | Var _ as e -> e
      | Let(letty, (false as isrec), qvars, bindings, e) ->
         Let(letty, isrec, qvars, List.map (fun (x, sc, e) -> (x, sc, un e)) bindings, un e)
      | MultiFix(bindings, ret) ->
         MultiFix(List.map (fun (a, b, e) -> (a, b, un e)) bindings, ret)
      | Match(e,clauses) ->
         Match(un e, List.map (fun (p, e) -> (p, un e)) clauses)
      | ShallowMatch(v, clauses) ->
         ShallowMatch(v, List.map (fun (c, vs, e) -> (c, vs, un e)) clauses)
      | Con(c, ts, es) ->
         Con(c, ts, List.map un es)
      | ShallowCon _ as e -> e
      | Abs(x, t, e) -> Abs(x, t, un e)
      | App(e1, e2) -> App(un e1, un e2)
      | Mark(m, e) -> Mark(m, un e)
      | Let(letty, true, qvars, bindings,e) ->
         (* Put each definition in its own multifix ! That requires renaming the let a few times *)
         let let_tail ((_, tail_ty, _) as tail) =
           Let(letty, true, qvars, bindings,
               tail), tail_ty, loc
         in
         let bindings =
           List.map (fun (x,Scheme(tyvars,ty),e) ->
             let rn_let = SubstRename.fresh_term (let_tail (Var(x, []), ty, loc)) in
             let e_unrec = match rn_let with
               | Let(_,_,_,bindings, (Var(x, []),_,_)),_,_ ->
                  let bindings =
                    List.map (fun (y, Scheme(y_tvs,y_ty), y_e) ->
                      let (z, z_ty, e, e_ty) = extract_lambda y_e in
                      (y, (z,z_ty), e)
                    ) bindings
                  in
                  (MultiFix(bindings, x), ty, loc)
               | _ -> assert false
             in
             (x, Scheme(tyvars, ty), e_unrec)
           ) bindings
         in
         (* FIXME: this may make us remove some important stuff. we should probably have
            a better multifix to encode mutual recursion instead *)
         Let(IsMeta, false, qvars, bindings, e)
    in
    un

  let transform = unrecurse
end

module PolyLetExpansion = struct
  type mark += LetWasHere of id * loc | ExpandedFromLet of id * loc

  let instantiate (qvars, scvars, term) inst =
    let inst = List.combine scvars inst  in
    let instmap x =
      try List.assoc x inst
      with Not_found -> 
           if List.mem x qvars
           then TyProduct([])
           else TyVar(x)
    in
    SubstRename.(
      let (term_subst, _) = make_subst id_term instmap in
      term_subst term
    )

  let polyexpand =
    (* env : id -> (qvars, scvars, term)
       instantiate all qvars that are not in scvars with unit,
       ie. TyProduct []
     *)
    let rec exp (env : (id * (id list * id list * expr)) list) (e,ty,loc) = exp' env ty loc e
    and exp' env ty loc = function
      | Let(_, true, _, _, _) -> failwith "Called PolyLetExpansion on term with let rec"
      | MultiFix(bindings, ret) ->
         MultiFix(List.map (fun (a, b, e) -> (a, b, exp env e)) bindings, ret), ty, loc
      | Match(e,clauses) ->
         Match(exp env e, List.map (fun (p, e) -> (p, exp env e)) clauses), ty, loc
      | ShallowMatch(v, clauses) ->
         ShallowMatch(v, List.map (fun (c, vs, e) -> (c, vs, exp env e)) clauses), ty, loc
      | Con(c, ts, es) ->
         Con(c, ts, List.map (exp env) es), ty, loc
      | ShallowCon _ as e -> e, ty, loc
      | Abs(x, t, e) -> Abs(x, t, exp env e), ty, loc
      | App(e1, e2) -> App(exp env e1, exp env e2), ty, loc
      | Mark(m, e) -> Mark(m, exp env e), ty, loc
      | Var(x,inst) ->
         begin match List.assoc x env with
         | t -> instantiate t inst
         | exception Not_found ->
                     Var(x, inst), ty, loc
         end
      | Let(letty, (false as isrec), ([] as qvars), bindings, e) ->
         (* Monomorphic let *)
         Let(letty, isrec, qvars,
             List.map (fun (x,sc,e) -> (x, sc, exp env e)) bindings,
             exp env e), ty, loc
      | Let(letty, false, qvars, bindings, e) ->
         let let_id = id "expanded_let" in
         let env_bindings =
           List.map (fun (x,Scheme(scvars,_),e) ->
                     (* If bindings bind a var, we have to memorize them *)
                     register_expr_subst x e;
                     (x,(qvars, scvars, mark (ExpandedFromLet(let_id,loc)) (exp env e)))
                    ) bindings
         in
         mark (LetWasHere(let_id, loc)) (exp (env_bindings@env) e)
    in
    exp []

  let transform = polyexpand
end

module ANFTransform = struct
  (* c(a1,...,an) -> let xn = an and ... x1 = a1 in c(x1,...,xn) ;
     match a with ... -> let x = a in match x with ...
     In the case of recursive ANF, try to put all the lets at the same level
     (allows more equalities)
   *)


  let anf =
    let rec anf lv (e,ty,loc) = (anf' lv ty loc e, ty, loc)
    and anf' lv ty loc = function
      | Var(x, inst) -> Var(x, inst)
      | MultiFix(bindings, ret) ->
         MultiFix(List.map (fun (a, b, e) -> (a, b, anf lv e)) bindings, ret)
      | ShallowMatch(v, clauses) ->
         ShallowMatch(v, List.map (fun (c, vs, e) -> (c, vs, anf lv e)) clauses)
      | ShallowCon _ as e -> e
      | Abs(x, t, e) -> Abs(x, t, anf lv e)
      | App(e1, e2) -> App(anf lv e1, anf lv e2)
      | Mark(m, e) -> Mark(m, anf lv e)
      | (Let _ | Match _ | Con _ ) as e ->
         let (bindings, e) = anf_collect lv (e,ty,loc) in
         (* Folding over the bindings is necessary because they can depend on each other *)
         let e,_,_ = List.fold_right (fun (is_meta,x,((_,ty,loc) as f)) e ->
                                      Let(is_meta, false, [], [(x,Scheme([],ty),f)], e), ty, loc) bindings e
         in
         e
    and anf_collect lv (e,ty,loc) = match e with
      | Let(is_meta, false, [], bindings, e) ->
         let bss = List.map (fun (x, _, e) -> let bs, e = anf_collect lv e in bs@[(is_meta,x,e)] ) bindings in
         let bs', e = anf_collect lv e in
         (List.concat bss @ bs'), e
      | Match((_, e_ty', e_loc') as e', clauses) ->
         let x = id "x" in
         let s = "_" ^string_of_int lv in
         let mx = id s in
         let (bs, e') = anf_collect lv e' in
         let e' = if Eml_untyped.is_expansive e' then (say "Marked one term %d@." lv; Mark(NiceName s, e'), e_ty', e_loc') else e' in
         bs@[(IsShare,x,e');(IsShare,mx, (Var(x,[]),ty,loc))], (Match((Var(x,[]), e_ty', dummy_loc) ,
                             List.map (fun (p,e) -> (p, anf (lv+1) e)) clauses), ty, loc)
      | Con(c,tys,es) ->
         let xs = List.map (fun _ -> id "x") es in
         let bss = List.map2 (fun x e -> let (bs, e) = anf_collect lv e in
                                         bs@[(IsShare,x,e)]
                             ) (List.rev xs) (List.rev es) in
         List.concat bss, (ShallowCon(c,tys,xs,id "cons"), ty, loc)
      | _ -> [], anf lv (e,ty,loc)
    in
    anf 1

  let transform = anf

end

module PatternSimplification = struct
  type mark += SimplifiedMatch of id * loc
  exception OverlappingPatterns
  exception IncompletePatterns

  type mode =
    | NoChecks
    | CheckComplete
    | CheckPartition

  let rec empty_or_wild_pattern (p,_,_) = match p with
    | PZero -> true
    | PAlt(p1,p2) -> empty_or_wild_pattern p1 || empty_or_wild_pattern p2
    | _ -> false

  let match_unfold check_overlap subtransform strict_typing env match_id ty loc =
    let rec unfold vars all_clauses =
      match all_clauses with
      | [] -> raise IncompletePatterns
      | (pat,act)::clauses ->
         (* Assumption: everything seen before is a var pattern *)
         let rec scan leftv leftpv rightv rightp =
           match rightv, rightp with
           | [], [] ->
           (* Generate a bunch of let, rename *)
              let term =
                Let(IsShare, false, [],
                  List.concat (List.map2 (fun (x,x_ty) ->
                                          function
                                          | PWild, _, _ -> []
                                          | PVar y, _, _ -> [(y, Scheme([],x_ty), (Var(x,[]), x_ty, dummy_loc))]
                                          | _ -> assert false
                                         ) leftv leftpv),
                  subtransform act), ty, loc
              in
              if not (List.for_all (fun (ps,_) -> List.exists empty_or_wild_pattern ps) clauses) && check_overlap then raise OverlappingPatterns;
              SubstRename.fresh_term term
           | (x::xs), (((PVar _ | PWild),_,_) as p)::ps ->
              scan (x::leftv) (p::leftpv) xs ps
           | (x::xs), (PAlt(p1, p2),_,_)::ps ->
              unfold vars ((List.rev leftpv @ [p1] @ ps, act)::
                           (List.rev leftpv @ [p2] @ ps, act)::
                             clauses)
           (* Possible improvement: Early cancel on PZero *)
           | (x::xs), (PZero,_,_)::ps ->
              unfold vars clauses
           | (x::xs), (PCon(c,_,_),ty,_)::ps ->
              specialize_constructor c (List.length leftv) vars all_clauses
           | _ -> assert false
         in
         scan [] [] vars pat
    and specialize_constructor c idx vars clauses =
      (* Iterate on constructors *)
      let (lv, (x,x_ty), rv) = cut_list idx vars in
      let constructors =
        (* Compute the constructors and the types of the arguments inside *)
        let (ty_head, ty_args) =
          match x_ty with
          | TyCon(ty_head, ty_args) when strict_typing -> (ty_head, ty_args)
          | _ when strict_typing -> assert false
          | _ ->
             let (ty_head, ty_args, _) = List.assoc c env.conenv in
             (ty_head, List.map (fun x -> TyVar(x)) ty_args)
        in
        let (x_args, conses) = List.assoc ty_head env.dataenv in
        let subst = List.combine x_args ty_args in
        let (_, instance) = SubstRename.(make_subst id_term (fun tv -> List.assoc tv subst)) in
        List.map (fun (c, args_ty) -> (c, List.map instance args_ty)) conses
      in
      let rec spec c xis (pats, act) =
        let (lp, p, rp) = cut_list idx pats in
        match p with
        | PZero, _, _ -> []
        | PAlt(p1, p2), _, _ ->
           spec c xis (lp@[p1]@rp,act) @ spec c xis (lp@[p2]@rp,act)
        | PWild, _, _ -> [(lp @ List.map (fun (_, ty) -> (PWild, ty, dummy_loc)) xis @ rp, act)]
        | PVar v, v_ty, _ ->
           [(lp@List.map (fun (_, ty) -> (PWild, ty, dummy_loc)) xis @ rp,
             (Let(IsShare, false, [], [v, Scheme([], v_ty), (Var(x,[]), x_ty, dummy_loc)], act), ty, loc)
            )]
        | PCon(c', _, pis), _, _ ->
           if c = c'
           then (
               assert (List.length pis = List.length xis);
               [(lp@pis@rp),act]
           ) else []
      in
      let outer_match =
        let branches =
          List.map
            (fun (c,argtypes) ->
             let xis = List.map (fun t -> (id "x", t)) argtypes in
             (c, List.map fst xis, unfold (lv@xis@rv) (List.concat (List.map (spec c xis) clauses)))
            ) constructors
        in
        ShallowMatch(x, branches), ty, loc
      in
      Mark(SimplifiedMatch(match_id, loc), outer_match), ty, loc
    in
    unfold

  (* This one requires a bit more thought *)
  let rec match_elim mode check_mode dataenv =
    let rec un (e,ty,loc) = (un' ty loc e, ty, loc)
    and un' ty loc = function
      | Var _ as e -> e
      | Let(letty, isrec, qvars, bindings, e) ->
         Let(letty, isrec, qvars, List.map (fun (x, sc, e) -> (x, sc, un e)) bindings, un e)
      | MultiFix(bindings, ret) ->
         MultiFix(List.map (fun (a, b, e) -> (a, b, un e)) bindings, ret)
      | ShallowMatch(v, clauses) ->
         ShallowMatch(v, List.map (fun (c, vs, e) -> (c, vs, un e)) clauses)
      | Con(c, ts, es) ->
         Con(c, ts, List.map un es)
      | ShallowCon _ as e -> e
      | Abs(x, t, e) -> Abs(x, t, un e)
      | App(e1, e2) -> App(un e1, un e2)
      | Mark(m, e) -> Mark(m, un e)
      | Match(e,clauses) ->
         let x,x_ty = match e with
           | Var(x,[]),x_ty,_ -> (x,x_ty)
           | _ -> failwith "PatternSimplification: expr in match"
         in
         let match_id = id "match" in
         let default_var = id "bad" in
         IdHashtbl.add Pretty.holemap default_var ("[incomplete pattern matching at "^string_of_loc loc^"]");
         let default_case = [([PWild, x_ty, dummy_loc], (Var(default_var,[]), TyVar(id "bad"), dummy_loc))] in
         let e,_,_ = match_unfold (check_mode = CheckPartition) (match_elim mode check_mode dataenv) mode dataenv match_id ty loc [(x,x_ty)] ((List.map (fun (p,e) -> ([p], e)) clauses)@ (if check_mode = NoChecks then default_case else [])) in
         e
    in
    un

  let transform env = match_elim true CheckComplete env
  let transform_untyped env d = match_elim false d env
end

let do_all env e =
  let e = LetUnrecursion.transform e in
  let e = PolyLetExpansion.transform e in
  let e = ANFTransform.transform e in
  let e = PatternSimplification.transform env e in
  e

let untyped_all env e =
  debug "%a@." Pp.term e;
  let e = LetUnrecursion.transform e in
  debug "%a@." Pp.term e;
  let e = PolyLetExpansion.transform e in
  debug "%a@." Pp.term e;
  let e = ANFTransform.transform e in
  debug "%a@." Pp.term e;
  let e = PatternSimplification.(transform_untyped env CheckComplete e) in
  debug "%a@." Pp.term e;
  e

let untyped_all_patch env e =
    debug "%a@." Pp.term e;
  let e = LetUnrecursion.transform e in
  debug "%a@." Pp.term e;
  let e = PolyLetExpansion.transform e in
  debug "%a@." Pp.term e;
  let e = ANFTransform.transform e in
  debug "%a@." Pp.term e;
  let e = PatternSimplification.(transform_untyped env NoChecks e) in
  debug "%a@." Pp.term e;
  e


let untyped_all_ornament env e =
  debug "%a@." Pp.term e;
  let e = LetUnrecursion.transform e in
  debug "%a@." Pp.term e;
  let e = PolyLetExpansion.transform e in
  debug "%a@." Pp.term e;
  let e = ANFTransform.transform e in
  debug "%a@." Pp.term e;
  let e = PatternSimplification.(transform_untyped env CheckPartition e) in
  debug "%a@." Pp.term e;
  e
