open Eml

module GmlReduction = struct
  (* A module to do GML reduction. It is capable of reducing open terms, so we can see
     what the user needs adding/what we can infer because it is trivial/what disappears
     without too much problems *)

  (* Make a copy of a term with all bound variables renamed.
     More convenient with mutable state, and safe since all bound identifier in
     a term are guaranteed to be distinct.
   *)
  open ExplicitGml
  let make_subst subst_te subst_ty =
    let open List in
    let h = IdHashtbl.create 17 in
    let bind x = let x' = fresh_id x in
                 IdHashtbl.add h x x'; x' in
    let rn_var x = try IdHashtbl.find h x with Not_found -> assert false in
    let rn_te x = try Var(IdHashtbl.find h x) with Not_found -> subst_te x in
    let rn_ty x = try TVar(IdHashtbl.find h x) with Not_found -> subst_ty x in
    let rec rte_ = function
      | Var(x) -> rn_te x
      | Label(s, e) -> Label(s, rte e)
      | Let(is_meta,bindings,e) ->
         let bindings = map (fun (x,ty,e) -> let x = bind x in (x, rty ty, rte e)) bindings in
         Let(is_meta,bindings,rte e)
      | Match(e,ps) ->
         let ps = map (fun (c,xs,e) -> let xs = map bind xs in (c,xs,rte e)) ps in
         Match(rte e, ps)
      | Con(c,ts,es) -> Con(c, map rty ts, map rte es)
      | Tuple(ts, es) -> Tuple(map rty ts, map rte es)
      | LetTuple(fields, e, f) ->
         let fields = List.map (fun (x,t) -> bind x, rty t) fields in
         LetTuple(fields, rte e, rte f)
      | Abs(x,ty,e) -> let x = bind x in Abs(x, rty ty, rte e)
      | App(e1,e2) -> App(rte e1, rte e2)
      | GAbs(x, e) -> let x = bind x in GAbs(x, rte e)
      | GApp(e1, e2) -> GApp(rte e1, rte e2)
      | GAbsT(x,e) -> let x = bind x in GAbsT(x, rte e)
      | GAppT(e,ty) -> GAppT(rte e, rty ty)
      | MultiFix(bindings, returned) ->
           List.iter (fun (x,_,_) -> ignore (bind x)) bindings;
           MultiFix(List.map (fun (x,(a,t),e) ->
                              let a = bind a in
                              (rn_var x, (a, rty t), rte e)
                             ) bindings,
                    rn_var returned)
    and rte (e,ty,loc) = (rte_ e, ty, loc)
    and rty  = function
      | TFall(x,ty) -> let x = bind x in TFall(x, rty ty)
      | TGFall(x,ty) -> let x = bind x in TGFall(x, rty ty)
      | TMatch(e, ps) ->
         let ps = map (fun (c,xs,ty) -> let xs = map bind xs in (c,xs,rty ty)) ps in
         TMatch(rte e, ps)
      | TCon(c,tys) -> TCon(c,map rty tys)
      | TVar(v) -> rn_ty v
      | TArr(t1,t2) -> TArr(rty t1, rty t2)
      | TGArr(x,t1,t2) -> let x = bind x in TGArr(x, rty t1, rty t2)
      | TAbs(x,t1,t2) -> let x = bind x in TAbs(x, rty t1, rty t2)
      | TApp(t,e) -> TApp(rty t, rte e)
      | TAbsT(x, t) -> let x = bind x in TAbsT(x, rty t)
      | TAppT(t1, t2) -> TAppT(rty t1, rty t2)
      | TProd(ts) -> TProd(List.map rty ts)
    in
    (rte, rty)
  let fresh_term te = let (rte,_) = make_subst (fun x -> Var(x)) (fun x -> TVar x) in rte te
  let fresh_type ty = let (_,rty) = make_subst (fun x -> Var(x)) (fun x -> TVar x) in rty ty

  (* Do as much reduction as possible. With an environment!
     Invariant: everything in the env is maximally (head-)reduced
   *)
  let rec reduce_te_aux ty_env te_env (e,ty,loc) =
    let rty = reduce_ty ty_env te_env in
    let rte = reduce_te ty_env te_env in
    let oty = rty ty in
    match e with
    | Var x  ->
       begin match List.assoc x te_env with
             | e -> fresh_term e
             | exception Not_found -> (e,oty,loc)
       end
    | Label(s, e) -> Label(s, rte e), oty, loc
    | Let(is_meta,bindings,e) ->
       let bindings = List.map (fun (kk,ty,e) -> (kk,rty ty, rte e)) bindings in
       Let(is_meta,bindings,rte e), oty, loc
    | MultiFix(bindings, ret) ->
       let bindings = List.map (fun (a,(b,t),e) -> (a,(b,rty ty),rte e)) bindings in
       MultiFix(bindings, ret), oty, loc
    | Match(e,ps) ->
       Match(rte e, List.map (fun (c,xs,e) -> (c,xs,rte e)) ps), oty, loc
    | Con(c,tys,es) ->
       Con(c,List.map rty tys, List.map rte es), oty, loc
    | Tuple(ts, es) -> Tuple(List.map rty ts, List.map rte es), oty, loc
    | LetTuple(flds, e, f) -> LetTuple(List.map (fun (x,t) -> (x,rty t)) flds, rte e, rte f), oty, loc
    | Abs(x,ty,e) -> Abs(x, rty ty, rte e), oty, loc
    | App(e1,e2) -> App(rte e1, rte e2), oty, loc
    | GAbs(x,e) -> GAbs(x,rte e), oty, loc
    | GAbsT(x,e) -> GAbsT(x, rte e), oty, loc
    | GApp(e1, e2) ->
       let e2 = rte e2 in
       begin match rte e1 with
             | GAbs(x,e), _, _ -> reduce_te ty_env ((x,e2)::te_env) e
             | e1 -> GApp(e1, e2), oty, loc
       end
    | GAppT(e,ty) ->
       let ty = rty ty in
       begin match rte e with
             | GAbs(x,e), _, _ -> reduce_te ((x,ty)::ty_env) te_env e
             | e -> GAppT(e,ty), oty, loc
       end
  and reduce_te ty_env te_env (e,ty,loc) =
    let r = reduce_te_aux ty_env te_env (e,ty,loc) in
    (*     Format.eprintf "Reduced %a to %a@." Pp.eml_expr' e Pp.eml_expr r; *)
    r
  and reduce_ty ty_env te_env ty =
    let rte = reduce_te ty_env te_env in
    let rty = reduce_ty ty_env te_env in
    match ty with
    | TProd(ts) -> TProd(List.map rty ts)
    | TFall(x, ty) -> TFall(x, rty ty)
    | TGFall(x,ty) -> TGFall(x, rty ty)
    | TMatch(e, ps) -> TMatch(rte e, List.map (fun (c,xs,ty) -> (c,xs,rty ty)) ps)
    | TCon(c, tys) -> TCon(c, List.map rty tys)
    | TVar(x) -> (try fresh_type (List.assoc x ty_env) with Not_found -> TVar(x))
    | TArr(t1,t2) -> TArr(rty t1, rty t2)
    | TGArr(x,t1,t2) -> TGArr(x,rty t1, rty t2)
    | TAbs(x, t1, t2) -> TAbs(x, rty t1, rty t2)
    | TAbsT(x, t) -> TAbsT(x, rty t)
    | TApp(t,e) ->
       let e = rte e in
       begin match rty t with
             | TAbs(x,t1,t2) -> reduce_ty ty_env ((x,e)::te_env) t2
             | t -> TApp(t,e)
       end
    | TAppT(t1,t2) ->
       let t2 = rty t2 in
       begin match rty t1 with
             | TAbsT(x,t) -> reduce_ty ((x,t2)::ty_env) te_env t
             | t1 -> TAppT(t1,t2)
       end
end

module EmlToImplicit = struct
  (* It's supposedly converted to the typed language, but all typing information is discarded *)
  (* FIXME: unshare the type *)

  module E = ExplicitGml
  module X = Ast.Explicit
  module A = Ast
  let ty = A.TyVar(id "untyped")
  let sc = A.Scheme([], ty)

  (* Somewhere we get the list of instantiation things. This makes it easier to track holes in
     the term. Here we propagate the hole information from gML down to ML
   *)
  let analyze_hole e = assert false

  let rec term_to_ml' = function
    | E.Var(x) -> X.Var(x,[])
    | E.Let(is_meta,bindings, e) -> X.Let(is_meta, false, [], List.map (fun (x,_,e) -> (x,sc,cv e)) bindings, cv e)
    | E.Tuple(ts,es) -> X.Con(tuple_con.(List.length es), List.map (fun _ -> ty) ts, List.map cv es)
    | E.LetTuple(ys,e,f) ->
       let x = id "x" in
       X.Let(IsMeta, false, [], [(x,sc,cv e)],
             (X.ShallowMatch(x, [tuple_con.(List.length ys), List.map fst ys, cv e]), ty, dummy_loc))
    | E.MultiFix(bs, x) -> X.MultiFix(List.map (fun (x,(y,_),e) -> (x,(y,ty),cv e)) bs,x)
    | E.Match(e, cls) ->
       let x = id "x" in
       X.Let(IsUser, false, [], [(x,sc,cv e)],
             (X.ShallowMatch(x, List.map (fun (x,vs,e) -> (x,vs,cv e)) cls), ty, dummy_loc))
    | E.Con(c,_,es) ->
       let bindings = List.map (fun e -> (id "x", sc, cv e)) es in
       X.Let(IsUser, false, [], bindings,
             ( X.ShallowCon(c, [], List.map (fun (x,_,_) -> x) bindings, id "unloc_cons"), ty, dummy_loc ))
    | E.Abs(x,_,e) -> X.Abs(x,ty,cv e)
    | E.App(e1,e2) -> X.App(cv e1, cv e2)
    (* This is cheating *)
    | E.GApp(e1,e2) -> X.App(cv e1, cv e2)
    | E.Label(s,e) -> X.Mark(X.NiceName s, cv e)
    | e -> failwith "Unhandled construction in conversion to eML"
  and cv (x,_,loc) = term_to_ml' x, ty, loc

end

module ExplicitToGml = struct
  (* This is bad with the types... *)
  module E = ExplicitGml
  module X = Ast.Explicit
  module A = Ast

  let dummy_ty = E.TVar(id "#INVALID#")
  let ml_to_gml =
    let cv_var loc x = E.Var(x), dummy_ty, loc in
    let rec cv_aux loc = function
      | X.Var(x, ts) -> List.fold_left (fun e t -> failwith "that should have disappeared by now") (E.Var x) ts
      | X.Let(_, true, _, _, _) -> failwith "recursive let should not appear here"
      | X.Let(_, false, _::_,_,_) -> failwith "polymorphic let should not appear here"
      | X.Let(is_meta, false, [], bs, e) ->
         E.Let(is_meta, List.map (fun (x, A.Scheme(tvs,ty), e) -> assert (tvs = []); (x, E.of_ml_ty ty, cv e)) bs, cv e)
      | X.MultiFix(bs, x) -> E.MultiFix(List.map (fun (x,(y,ty),e) -> (x,(y,E.of_ml_ty ty), cv e)) bs, x)
      | X.Match(_,_) -> failwith "full pattern matching should not appear here"
      | X.ShallowMatch(x, cls) ->
         E.Match(cv_var loc x,
                 List.map (fun (c,xs,e) -> (c, xs, cv e)) cls)
      | X.Con(c,ts,es) -> E.Con(c, List.map E.of_ml_ty ts, List.map cv  es)
      | X.ShallowCon(c,ts,xs,_conid) -> E.Con(c, List.map E.of_ml_ty ts, List.map (cv_var loc)  xs)
      | X.Abs(x,ty,e) -> E.Abs(x, E.of_ml_ty ty, cv e)
      | X.App(e, f) -> E.App(cv e, cv f)
      | X.Mark(X.NiceName s, e) -> Label(s, cv e)
      | X.Mark(_,e) -> let (e, _, _) = cv e in e
    and cv (x,ty,loc) = (cv_aux loc x, E.of_ml_ty ty, loc)
    in
    cv
end
