
(* Here we do match-reduction and we desugar let-recs. Then, we retype
   our term to check that it is ok
 *)

(* Match reduction:
    walk the term, keeping in mind the equalities we obtain from "let" and "match"
    duplicate when needed (ie let match).
 *)
open Ast
open Util
open! Ast.Explicit


(* This is very partial *)
exception Not_unifiable
let unify_terms t1 t2 =
  let unif = ref [] in
  let bound = ref [] in
  let uni_var v1 v2 =
    if (List.mem v1 !bound || List.mem v2 !bound) then
      (if not (List.mem (v1,v2) !unif) then raise Not_unifiable)
    else unif := (v1,v2) :: !unif
  in
  let uni_bind v1 v2 =
    uni_var v1 v2;
    bound := v1::v2::!bound
  in
  let rec uni (e1,_,_) (e2,_,_) =
    match e1, e2 with
    | Var(x1,_), Var(x2,_) -> uni_var x1 x2
    | App(e1, f1), App(e2, f2) -> uni e1 e2; uni f1 f2
    | Abs(x1,_,e1), Abs(x2,_,e2) -> uni_bind x1 x2; uni e1 e2
    | MultiFix(bs1, x1), MultiFix(bs2, x2) ->
       uni_var x1 x2;
       List.iter2 (fun (x1, (y1,_), e1) (x2, (y2,_), e2) -> uni_bind x1 x2; uni_bind y1 y2; uni e1 e2) bs1 bs2
    | Let(_,_,_,bs1,e1), Let(_,_,_,bs2,e2) ->
       uni e1 e2; List.iter2 (fun (x1,_,e1) (x2,_,e2) -> uni_bind x1 x2; uni e1 e2) bs1 bs2
    | ShallowCon(c1,_,xs1,_), ShallowCon(c2,_,xs2,_) when c1 = c2 -> List.iter2 uni_var xs1 xs2
    | ShallowMatch(x1,cls1), ShallowMatch(x2,cls2) ->
       uni_var x1 x2;
       List.iter2
         (fun (c1,xs1,e1) (c2,xs2,e2) ->
          if c1 <> c2 then raise Not_unifiable;
          List.iter2 uni_bind xs1 xs2;
          uni e1 e2
         ) cls1 cls2
    | Mark(m1, e1), Mark(m2, e2) -> uni e1 e2 (* ? *)
    | Match(e1, cls1), Match(e2, cls2) ->
       uni e1 e2; List.iter2 (fun (p1,e1) (p2,e2) -> uni_pat p1 p2; uni e1 e2) cls1 cls2
    | Con(c1,_,es1), Con(c2,_,es2) when c1 = c2 -> List.iter2 uni es1 es2
    | _ -> raise Not_unifiable
  and uni_pat (p1,_,_) (p2,_,_) =
    match p1, p2 with
    | PVar(x1), PVar(x2) -> uni_bind x1 x2
    | PCon(c1,_,ps1), PCon(c2,_,ps2) when c1 = c2 -> List.iter2 uni_pat ps1 ps2
    | PWild, PWild -> ()
    | PZero, PZero -> ()
    | PAlt(p1,q1), PAlt(p2,q2) -> uni_pat p1 p2; uni_pat q1 q2
    | _ -> raise Not_unifiable
  in
  try
    (* XXX uni_var is strange. *)
    uni t1 t2; Some(!unif)
  with
    Not_unifiable -> None

let ty = TyVar(id "untyped")
let dummy_ty = ty
let sc = Scheme([], ty)

module Context = struct
  (* This should be a persistent union-find *)
  type ctx = (id list * (id * id list) option) list

  let get_entry ctx u=
    match List.filter (fun (l,_) -> List.mem u l) ctx with
    | [] -> ([u], None)
    | [e] -> e
    | _ -> assert false

  let get_head ctx u =
    snd (get_entry ctx u)

  let remove_entry ctx u =
    List.filter (fun (l,_) -> not (List.mem u l)) ctx

  (* v is old, u is new *)
  let push_eq ctx v u =
    let (vv,hv) = get_entry ctx v in
    let ctx = remove_entry ctx v in
    (u::vv,hv)::ctx

  (* u is old, the vs are new. u has no head *)
  let push_head ctx u c vs =
    let (vu,None) = get_entry ctx u in
    let ctx = remove_entry ctx u in
    (vu, Some(c,vs))::ctx

  let tuple_head = id "(t,u,p,l,e)"
end
module C = Context

(* Match reduction should track what happens to holes.
   We insert 'let' instead of doing substitution, so the
   disappearing-names problems can be solved later *)
let rec match_reduction ctx = function
  | ((Var _ | ShallowCon _),_,_) as e -> e
  | Mark(NiceName n,e), ty, loc -> Mark(NiceName n, match_reduction ctx e), ty, loc
  | (ShallowMatch(x, cls), ty, loc) as e ->
     begin match C.get_head ctx x with
           | None ->
              ShallowMatch(x, List.map (fun (c,vs,e) ->
                                        (c,vs, match_reduction (C.push_head ctx x c vs) e) 
                                       ) cls), ty, loc
           | Some(c,vs) ->
              (* Select a branch, insert the equalities *)
              begin match List.filter (fun (c',_,_) -> c = c') cls with
                  | [(_,vs',e)] ->
                     Let(IsShare, false, [], List.map2 (fun vnew vold -> (vnew, sc, (Var(vold,[]),ty,dummy_loc))) vs' vs,
                         match_reduction (List.fold_left2 C.push_eq ctx vs vs') e), ty, loc
                  | _ -> Format.eprintf "When treating:\n%a\nSelected branch %a@." Pp.xterm e Pp.id c; failwith "type error in patches"
              end
     end
  | App(e1, e2), ty, loc -> App(match_reduction ctx e1, match_reduction ctx e2), ty, loc
  | Abs(x,tx,e), ty, loc -> Abs(x,tx,match_reduction ctx e), ty, loc
  | MultiFix(bs, x), ty, loc -> MultiFix(List.map (fun (a,b,e) -> (a,b,match_reduction ctx e)) bs, x), ty, loc
  | Let(is_meta,false,[],bs,e), ty, loc ->
     treat_let is_meta ctx e bs
  | _ -> failwith "Unhandled syntax in match_reduction"
and treat_let is_meta ctx e = function
  | [] -> match_reduction ctx e
  | (x,_, (Var(y,_),_,_)) as b::bs ->
     let ctx = C.push_eq ctx x y in
     Let(is_meta,false,[],[b], treat_let is_meta ctx e bs), ty, dummy_loc
  | (x,_,(ShallowCon(c,_,vs,_),_,_)) as b::bs ->
     let ctx = C.push_head ctx x c vs in
     Let(is_meta, false, [], [b], treat_let is_meta ctx e bs), ty, dummy_loc
  | (x,_, (ShallowMatch(y,cls),ty,loc)) :: bs ->
     match_reduction ctx (
                       ShallowMatch(y, List.map (fun (c,vs,e') ->
                                                 (c,vs,  SubstRename.fresh_term (( Let(is_meta, false, [], (x,sc,e')::bs, e),ty,loc )) )) cls), ty, dummy_loc
                     )
  | b :: bs ->
     let (x,sc,f) = b in
     Let(is_meta, false, [], [(x,sc,match_reduction ctx f)], treat_let is_meta ctx e bs), ty, dummy_loc

let rec is_expansive (e,_,_) = match e with
  | (Var _ | Abs _ | MultiFix _ | ShallowCon _) -> false
  | App _ -> true
  | Con(_,_,es) -> List.exists is_expansive es
  | ShallowMatch(_, cls) -> List.exists (fun (_,_,e) -> is_expansive e) cls
  | Let(_,_,_,bs,e) -> is_expansive e || List.exists (fun (_,_,e) -> is_expansive e) bs
  | Mark(NiceName _,_) -> true
  | _ -> true

(* Finally, we resugar a bit. A large set of possible transformations:
   + move linear lets when this does not change the evaluation order
   + substitute variable = variable lets
   + replace unused variables by _
   + replace stuff by let-rec
   + merge multiple lets
 *)
let move_nonexpansive e =
  let rec eval env (e,ty,loc) = match e with
    | Mark(NiceName s,e) -> Mark(NiceName s, eval env e), ty, loc
    | Var(x,_) -> (try List.assoc x env with _ -> (e,ty,loc))
    | App(e1, e2) -> App(eval env e1, eval env e2), ty, loc
    | Abs(x,t,e) -> Abs(x,t, eval env e), ty, loc
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,eval env e)) bs, x), ty, loc
    | Let(is_meta,is_rec,vs,bs,e) ->
       assert (not is_rec);
       let (bs_exp, bs_nonexp) = List.partition (fun (_,_,e) -> is_expansive e || is_meta = IsUser) bs in
       List.iter (fun (x,_,e) -> register_expr_subst x e) bs_nonexp;
       let e = eval (List.map (fun (x,_,e) -> (x,eval env e)) bs_nonexp@env) e in
       let bs_exp = List.map (fun (x,sc,e) -> (x,sc,eval env e)) bs_exp in
       if bs_exp = []
       then e
       else Let(is_meta, is_rec, vs, bs_exp, e), ty, loc
    | ShallowCon(c,ts,xs,_conid) -> Con(c,ts,List.map (eval_var env) xs), ty, loc
    | ShallowMatch(x,cls) ->
       begin match List.assoc x env with
             | Con(c,_,es), _, _ ->
                let (_,xs,e) = try List.find (fun (c',_,_) -> c = c') cls with Not_found -> failwith "Type error in patch" in
                eval env (Let(IsShare,false,[],List.map2 (fun x e -> (x,sc,SubstRename.fresh_term e)) xs es,e), ty, loc)
             | ShallowCon _, _, _ -> assert false (* Since the things in environment are evaluated, ShallowCon should not appear *)
             | e -> Match(SubstRename.fresh_term e, List.map (fun (c,xs,e) -> ((PCon(c,[],List.map (fun x -> PVar x, ty, loc) xs),ty,loc),eval env e)) cls), ty, loc
             | exception Not_found -> Match((Var(x,[]), ty, loc), List.map (fun (c,xs,e) -> ((PCon(c,[],List.map (fun x -> PVar x, ty, loc) xs),ty,loc),eval env e)) cls), ty, loc
       end
    | _ -> debug "Dropped %a@." Pp.term (e,ty,loc); (e,ty,loc)
  and eval_var env x =
    match List.assoc x env with
    | e -> e
    | exception Not_found -> Var(x, []), ty, dummy_loc
  in
  eval [] e


let rec pat_to_term (p,ty,loc) =
  match p with
  | PVar(x) -> Var(x,[]), ty, loc
  | PCon(c,ts,ps) -> Con(c, ts, List.map pat_to_term ps), ty, loc
  | _ -> failwith "pattern is not an expression"

(* The idea: e was matched with pat. now, we match x with cls.
   can "pat" be refined according to cls ? (ie transform it into a single match on e with pat')
   If yes, we return "Subst" and a list of clauses, otherwise "NoSubst" *)
type 'a subst_res = Subst of ('a * expr) list | NoSubst of 'a
let try_subst pat x cls =
  let combine f x y = match x,y with
    | NoSubst p1, NoSubst p2 -> NoSubst (f p1 p2)
    | Subst cls, NoSubst q -> Subst (List.map (fun (p,e) -> (f p q, e)) cls)
    | NoSubst p, Subst cls -> Subst (List.map (fun (q,e) -> (f p q, e)) cls)
    | Subst _, Subst _ -> failwith "Duplicate variable in pattern"
  in
  let map f = function
    | NoSubst p -> NoSubst (f p)
    | Subst cls -> Subst (List.map (fun (p,e) -> (f p, e)) cls)
  in
  let rec subst (p,ty,loc) = match p with
    | PVar x' when x = x' ->
       (* Be careful: we remove a binding of x', so we need to substitute it in the body *)
       Subst(List.map (fun (p,e) -> (p, SubstRename.(let (rte,_) = make_subst (fun x tys -> if x = x' then (assert (tys = []); let (ep',_,_) = pat_to_term p in ep') else Var(x,tys)) id_type in rte e))) cls)
    | (PVar _ | PWild | PZero) as p -> NoSubst (p, ty, loc)
    | PAlt(p1, p2) -> combine (fun p1 p2 -> PAlt(p1,p2), ty, loc) (subst p1) (subst p2)
    | PCon(c,ts,ps) ->
       map (fun ps -> PCon(c,ts,ps) , ty, loc)
           (List.fold_right (combine (fun x xs -> x::xs)) (List.map subst ps) (NoSubst []))
  in
  subst pat


(* Check if term variable x is free in e *)
exception Found
let occurs xs e =
  try
    let (rte,_,_) =
      SubstRename.(make_subst_full (fun x' tys -> if IdSet.mem x' xs then raise Found; id_term x' tys) id_type) in
    ignore (rte e);
    false
  with Found -> true

let add_wildcards (p,e) =
  let rec proc (p,ty,loc) = match p with
    | PVar x when not (occurs (IdSet.singleton x) e) -> PWild, ty, loc
    | PVar x -> PVar x, ty, loc
    | PCon(c,ts,ps) -> PCon(c,ts, List.map proc ps), ty, loc
    | _ -> failwith "Non-simple pattern"
  in
  (proc p, e)

let rec subst_pat subst (p,ty,loc) = match p with
  | PVar x -> PVar (List.assoc x subst), ty, loc
  | PWild | PZero -> (p,ty,loc)
  | PCon(c,ts,ps) -> PCon(c,ts,List.map (subst_pat subst) ps), ty, loc
  | PAlt(p1,p2) -> PAlt(subst_pat subst p1, subst_pat subst p2), ty, loc

let merge_clauses cls = 
  let cls =
    List.fold_left (fun cls (p,e) ->
      (* Can we merge the clause with any of our clauses *)
      match modify_one (fun ((_,ty,_) as p',e') ->
        let vs = patvars p in
        let vs' = patvars p' in 
        let subst_valid subst = 
          (* Should do more *)
          List.for_all (fun (v,v') -> v == v' || (List.mem v vs && List.mem v' vs')) subst
        in 
        match unify_terms e e' with
        | Some subst when subst_valid subst-> (* Subst goes from p-variables to p'-variables *)
           Some ((PAlt(p', subst_pat subst p), ty, dummy_loc), e')
        | _ -> None
      ) cls with
      | None -> (p,e)::cls
      | Some l -> l
    ) [] cls in
  List.rev cls

(* merge_match  & move_expansive *)
let merge_match e =
  let rec merge (e,ty,loc) = match e with
    | (ShallowMatch _ | ShallowCon _) -> assert false
    | Mark(NiceName s,e) -> Mark(NiceName s, merge e), ty, loc
    | Con(c,ts,es) -> Con(c,ts,List.map merge es), ty, loc
    | (Var _ as v) -> v, ty, loc
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,merge e)) bs, x), ty, loc
    | Let(is_meta, is_rec, vs, bs, e) ->
       Let(is_meta, is_rec, vs, List.map (fun (x,sc,e) -> (x,sc,merge e)) bs, merge e), ty, loc
    | App(e1,e2) -> App(merge e1, merge e2), ty, loc
    | Abs(x,t,e) -> Abs(x,t,merge e), ty, loc
    | Match(e, cls) ->
       (* Accumulate clauses, check they are on "legal" variables (should work well after
          match elimination *)
       Match(e, merge_clauses (List.concat (List.map accumulate cls))), ty, loc
    | _ -> debug "Dropped %a@." Pp.term (e,ty,loc); (e,ty,loc)
  and accumulate (p,(ee, _, _ as e)) =
    match ee with
    | Match((Var(x,[]),_,_), cls) ->
       (* Check if the variable is in p, compute the appropriate substitution *)
       begin match try_subst p x cls with
             | Subst(cls) -> List.concat (List.map accumulate cls)
             | NoSubst(_) -> [add_wildcards (p, merge e)]
       end
    | _ -> [add_wildcards (p, merge e)]
  in
  merge e

(* This should be replaced by a smarter mechanism using the same kind of labels to track effects.
   This would pass through match much better than the current hacks *)
type explist = Stop of id list | Cont of id list
let rec move_expansive e =
  (* We use uniqueness of variables to save some effort *)
  let moved = IdHashtbl.create 42 in
  let used = IdHashtbl.create 42 in
  let combine u v = match u, v with
    | Stop l, _ -> u
    | Cont l1, Cont l2 -> Cont (l1@l2)
    | Cont l1, Stop l2 -> Stop (l1@l2)
  in
  let stop = function
    | Stop l | Cont l -> Stop l
  in
  let scan x e el =
    let (Cont l | Stop l) = el in
    let rec scan = function
      | x'::xs when id_eq x x' -> IdHashtbl.add moved x e; Stop []
      | y::xs -> combine (Cont [y]) (scan xs)
      | [] -> raise Not_found
    in
    scan l
  in
  let rec list_vars (e,ty,loc) = match e with
    | (ShallowMatch _ | ShallowCon _) -> assert false
    | Var(x,_) ->
       let n = try IdHashtbl.find used x with Not_found -> 0 in
       IdHashtbl.replace used x (n+1);
       Cont [x]
    | Let(is_meta, is_rec, vs, bs, e) ->
       let ve = list_vars e in
       let l = match ve with |Stop l | Cont l -> l in
       (* We examine bindings in reverse, and bind any movable binding *)
       let (vbs,ve) = List.fold_right
                  (fun (x,_,f) (vbs,ve) ->
                   let uses = try IdHashtbl.find used x with Not_found -> 0 in
                   if uses = 1 && is_meta <> IsUser
                   then (match scan x f ve with
                         | ve -> ignore (list_vars f); vbs, ve
                         | exception Not_found -> (debug "Scan fail for %a" Pp.id x;
                                     combine (list_vars f) vbs, ve))
                   else combine (list_vars f) vbs, ve
                  ) bs (Cont [],ve) in
       combine vbs ve
    | Con(_,_,es) -> List.fold_right (fun e l -> combine (list_vars e) l) (List.rev es) (Cont [])
    | MultiFix(bs, x) -> List.iter (fun (_,_,e) -> ignore (list_vars e)) bs; Cont []
    | Match(e, [(p, f)]) ->
       combine (list_vars e) (list_vars f)
    | Match(e,cls) ->
       (* We don't handle branching well, but still make an effort to check whether
          the thing in the match is expansive or not (Cont / Stop) because it should
          allow a bit more *)
       let branches = List.map (fun (_,e) -> list_vars e) cls in
       let is_expansive = List.exists (function Stop _ -> true | _ -> false) branches in
       combine (list_vars e) (if is_expansive then Stop [] else Cont [])
    | App(e1,e2) -> stop (combine (list_vars e2) (list_vars e1))
    | Mark(NiceName s,e) -> list_vars e
    | Abs(x,t,e) -> ignore (list_vars e); Stop []
    | _ -> assert false
  in
  ignore (list_vars e);
  let rec move_defs (e,ty,loc) = match e with
    | Var(x,_) as v ->
       begin try move_defs (IdHashtbl.find moved x)
             with Not_found -> (v,ty,loc)
       end
    | (ShallowMatch _ | ShallowCon _) -> assert false
    | Con(c,ts,es) -> Con(c,ts,List.map move_defs es), ty, loc
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,move_defs e)) bs, x), ty, loc
    | Let(is_meta, is_rec, vs, bs, e) ->
       let bs =  map_filter (fun (x,sc,e) -> if IdHashtbl.mem moved x then None else Some(x,sc,move_defs e)) bs in
       if bs = [] then move_defs e
       else Let(is_meta, is_rec, vs, bs, move_defs e), ty, loc
    | App(e1,e2) -> App(move_defs e1, move_defs e2), ty, loc
    | Abs(x,t,e) -> Abs(x,t,move_defs e), ty, loc
    | Mark(NiceName s,e) -> Mark(NiceName s, move_defs e), ty, loc
    | Match(e, cls) ->
       let e = move_defs e in
       begin match e with
       | Con(c,_,es), _, _ ->
          (* Matches have already been merged so this might not work at all *)
          begin try
                  let ((PCon(_,_,ps), _, _), f) = List.find
                      (function
                      | (PCon(c', _, ps), _, _), _ -> c = c'
                       && List.for_all (function (PVar _), _ , _ -> true | _ -> false) ps
                      | _ -> false
                      ) cls in
                  let e' =
                    Let(IsMeta, false, [], List.rev (List.map2 (fun (PVar(x), ty, _) e -> (x, Scheme([], ty), e)) ps es), move_defs f),
                    ty, loc in
                  (* Re-process e' *)
                  move_expansive e'
            with
              Not_found ->
                Match(e, List.map (fun (p,e) -> p, move_defs e) cls), ty, loc
          end
       | _ ->
          (* Accumulate clauses, check they are on "legal" variables (should work well after
          match elimination *)
          Match(e, List.map (fun (p,e) -> p, move_defs e) cls), ty, loc
       end
    | _ -> warn "Dropped %a@." Pp.term (e,ty,loc); (e,ty,loc)
  in
  move_defs e


let elementary_simplification e =
  let rec simpl_aux env e = match e with
    | Var(x,l) -> Var(var env x, l)
    | App(e1, e2) -> App(simpl env e1, simpl env e2)
    | Abs(x,t,e) -> Abs(x,t,simpl env e)
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,simpl env e)) bs, x)
    | Let(a,b,c,bs,e) ->
       let bss, envs =
         List.split (
             List.map (fun (x,sc,e) ->
                       match simpl env e with
                       | Var(x',_), _, _ -> register_subst x x'; [], [(x,x')]
                       | e -> [(x,sc,e)],[]
                      ) bs
           ) in
       let bs, env = List.concat bss, List.concat envs @ env in
       if bs = [] then let(e,_,_) = simpl env e in e else Let(a,b,c,bs,simpl env e)
    | ShallowCon(c,ts,xs,conid) -> ShallowCon(c,ts, List.map (var env) xs,conid)
    | ShallowMatch(x,cls) ->
       ShallowMatch(var env x, List.map (fun (c,ys,e) -> (c,ys,simpl env e)) cls)
    | Mark(m,e) -> Mark(m, simpl env e)
    | _ -> debug "Dropped %a@." Pp.term (e,ty,dummy_loc); e
  and simpl env (e,ty,loc) = simpl_aux env e, ty, loc
  and var env x = (try List.assoc x env with Not_found -> x)
  in
  simpl [] e

    (*
(* Might be fragile for larger examples *)
let recursify_toplet (TopLet(x,sc,false,e) as tl) =

  match e with
  | MultiFix([(y,(z,ty),e)], y'), _, _ when y = y' ->
     TopLet(x,sc,true, (Abs(z,ty,SubstRename.(let (rte,_) = make_subst (fun x' tys -> if x' = y then (assert (tys = []); Var(x,[])) else Var(x',tys)) id_type in rte e)), dummy_ty, dummy_loc))
  | _ -> tl
    *)

exception HasHoles
let eliminate_labels e =
  let rec simpl_aux e = match e with
    | Var(x,_) when IdHashtbl.mem Pretty.holemap x -> raise HasHoles
    | Var(x,l) -> Var(x, l)
    | App(e1, e2) -> App(simpl e1, simpl e2)
    | Abs(x,t,e) -> Abs(x,t,simpl e)
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,simpl e)) bs, x)
    | Let(a,b,c,bs,e) ->
       Let(a,b,c, List.map (fun (x,sc,e) -> (x,sc,e)) bs, simpl e)
    | ShallowCon(c,ts,xs,conid) -> ShallowCon(c,ts,xs,conid)
    | ShallowMatch(x,cls) ->
       ShallowMatch(x, List.map (fun (c,ys,e) -> (c,ys,simpl e)) cls)
    | Mark(NiceName s, e) -> let (e,_,_) = simpl e in e
    | Mark(m,e) -> Mark(m, simpl e)
    | Match(e, cls) -> Match(simpl e, List.map (fun (p,e) -> (p, simpl e)) cls)
    | Con(c,ts,es) -> Con(c,ts,List.map simpl es)
  and simpl (e,ty,loc) = simpl_aux e, ty, loc
  in
  simpl e

let multifix_to_letrec e =
    let rec simpl_aux e = match e with
    | Var(x,l) -> Var(x, l)
    | App(e1, e2) -> App(simpl e1, simpl e2)
    | Abs(x,t,e) -> Abs(x,t,simpl e)
    | MultiFix(bs,x) -> MultiFix(List.map (fun (x,y,e) -> (x,y,simpl e)) bs, x)
    | Let(a,false,c,[(x,_,(MultiFix(bs,y),t,loc))],e) ->
       Let(a,true,c,
           List.map (fun (x,(y,t),e) -> (x,sc,(Abs(y,t,e),dummy_ty,dummy_loc))) bs,
           SubstRename.(let (rte,_) = make_subst (fun x' tys -> if x = x' then Var(y,tys) else Var(x',tys)) id_type in rte e))
    | Let(a,b,c,bs,e) ->
       Let(a,b,c, List.map (fun (x,sc,e) -> (x,sc,e)) bs, simpl e)
    | ShallowCon(c,ts,xs,conid) -> ShallowCon(c,ts,xs,conid)
    | ShallowMatch(x,cls) ->
       ShallowMatch(x, List.map (fun (c,ys,e) -> (c,ys,simpl e)) cls)
    | Mark(NiceName s, e) -> let (e,_,_) = simpl e in e
    | Mark(m,e) -> Mark(m, simpl e)
    | Match(e, cls) -> Match(simpl e, List.map (fun (p,e) -> (p, simpl e)) cls)
    | Con(c,ts,es) -> Con(c,ts,List.map simpl es)
  and simpl (e,ty,loc) = simpl_aux e, ty, loc
  in
  simpl e


(* BUG: multifix_to_letrec is buggy with mutual recursion *)
(* BUG: move_nonexpansive should remove unused recursive bindings *)

(* Inlining after unfolding:
  - non-user-specified liftings can be identified by having is_user = false set for their ids
  - we fold until we reach a definition we already folded
  - we use let bindings for the folding instead of doing substitution
*)

let inline_autodefs bindings =
  (* Acceptable substitutions: all auto-generated definitions
     They are identified by having a non-user name *)
  let subst = List.filter_map (fun (_,_,to_name,_,_,_,e) ->
      if to_name.id_user then None else Some (to_name, e)
    ) bindings in
  (* app_stack e [f1; f2; f3] -> e f1 f2 f3 *)
  let rec app_stack e = function
    | (f, ty, loc)::args -> app_stack (App(e, f), ty, loc) args
    | [] -> e
  in
  let rec inline already_inlined e =
    let rec inl (e, ty, loc) args =
      let re_app e = app_stack e args in
      match e with
      (* Attempt inlining *)
      | Var(x, _) when not (IdSet.mem x already_inlined) && List.mem_assoc x subst ->
         let inlined = inline (IdSet.add x already_inlined) (List.assoc x subst) in
         let inlined = SubstRename.fresh_term inlined in
         let rec app (e,ty,loc) args =
           (* XXX some things might be multifixes instead? *)
           match e, args with
           | Abs(x,ty_f,e), (f, ty, loc)::args ->
              Let(IsMeta, false, [], [(x, Scheme([], ty_f), f)], app e args), ty, loc
           | _ -> app_stack (e, ty, loc) args
         in
         app inlined args
      (* Accumulate arguments for application *)
      | App(e, f) -> inl e (((inl f []),ty,loc)::args)
      (* Map down the syntax tree *)
      | Var _ | ShallowCon _-> re_app (e, ty, loc)
      | Abs(x, t, e) -> re_app (Abs(x,t,inl e []), ty, loc)
      | MultiFix(bs, x) ->
         let bs = List.map (fun (x, (y, t), e) -> (x, (y, t), inl e [])) bs in
         re_app (MultiFix(bs, x), ty, loc)
      | Let(u,v,w,bs,e) ->
         let bs = List.map (fun (x, t, e) -> (x, t, inl e [])) bs in
         re_app (Let(u,v,w,bs,inl e []), ty, loc)
      | Match(e, cls) ->
         let cls = List.map (fun (p, e) -> (p, inl e [])) cls in
         re_app (Match(inl e [], cls), ty, loc)
      | ShallowMatch(x, cls) ->
         re_app (ShallowMatch(x, List.map (fun (c,xs,e) -> (c, xs, inl e [])) cls), ty, loc)
      | Con(c, ts, es) ->
         re_app (Con(c, ts, List.map (fun e -> inl e []) es), ty, loc)
      (* This may block some inlinings, do we want to keep marks on the stack instead? *)
      | Mark(m, e) -> re_app (Mark(m, inl e []), ty, loc)
    in
    inl e []
  in
  let bindings = List.map (fun (l,from_name,to_name,tvar_map,to_ov,genspec,t) ->
    (l, from_name, to_name, tvar_map, to_ov, genspec, inline (IdSet.singleton to_name) t)
  ) bindings in
  (* Determine the list of accessible bindings *)
  let accessible = IdHashtbl.create 5 in
  let rec dfs x =
    if not (IdHashtbl.mem accessible x)
    then begin
      IdHashtbl.add accessible x ();
      let (_,_,_,_,_,_,e) = List.find (fun (_,_,x',_,_,_,_) -> x = x') bindings in
      (* FIXME: this is quadratic *)
      List.iter (fun (_,_,y,_,_,_,_) ->
        if occurs (IdSet.singleton y) e
        then dfs y) bindings
    end
  in
  List.iter (fun (_,_,x,_,_,_,_) -> if x.id_user then dfs x) bindings;
  (* Keep only accessible bindings (iteratively) *)
  let bindings = List.filter (fun (_,_,x,_,_,_,_) -> IdHashtbl.mem accessible x) bindings in
  bindings

let resugar e = multifix_to_letrec (move_expansive (merge_match (move_nonexpansive e)))

let unrecursify_toplet (TopLet(is_rec, bs)) =
  if not is_rec then TopLet(false, bs)
  else
    let is_rec = not (List.for_all (fun (x,_,_) -> List.for_all (fun (_,_,e) -> not (occurs (IdSet.singleton x) e)) bs) bs) in
    TopLet(is_rec, bs)
