(* A prototype ! *)

open Error
open Id
open Parser
open Pp
open Pretty
open Ml_inference
open Desugaring
open Eml
open Eml_untyped
module OI = Orn_inference
open Ast
module X = Ast.Explicit
open Util
open Options


exception Patch_error of exn

(* I need to generalize everything in the initial env and put it in the gen_env... *)

let pretty_xterm ppf t =
  let t = PrettyNames.run_renamer ~warn:false (fun s ->
    Pretty.pretty_term s t
  ) in
  PPrint.ToChannel.pretty 1. 60 stderr (Pp.piterm Expr t);
  Format.fprintf ppf ""

(* Initialize the environment (and generalize it) with either, tuples *)
let initial_env =
  let initial_gen_env = Orn.{ skel_map = []; skel_unmap = [];  skel_conenv = []; skel_dataenv = []; gen_varenv = [] } in
  let initial_ml_env =
    let a,b = good_id "a", good_id "b" in
    { conenv = [(sum_l, (sum_ty,[a;b], [TyVar a]));
                (sum_r, (sum_ty,[a;b], [TyVar b]))] @
                 list_init max_tuple_size (fun i ->
                               let vars = Array.to_list (Array.init i (fun i -> good_id ("a"^string_of_int i))) in
                               tuple_con.(i), (tuple_ty.(i), vars, List.map (fun x -> TyVar x) vars));
      dataenv = [(sum_ty, ([a;b], [(sum_l, [TyVar a]);(sum_r, [TyVar b])]))] @
                  list_init max_tuple_size (fun i ->
                                let vars = Array.to_list (Array.init i (fun i -> good_id ("a"^string_of_int i))) in
                                tuple_ty.(i), (vars, [tuple_con.(i), List.map (fun x -> TyVar x) vars])
                               );
      varenv = [];
      scope = Scope.{empty_scope
                    with types = [(UserId "either", sum_ty)];
                         conses = [(UserId "Left", sum_l); (UserId "Right", sum_r)]
                    }
    }
  in
  let initial_defs =
    let a,b = good_id "a", good_id "b" in
    [true, (X.StType(X.TyDef(false,[sum_ty,[a;b], [(sum_l, [TyVar a]); (sum_r, [TyVar b])]])), dummy_loc)] @
      list_init 16 (fun i ->
                    let vars = Array.to_list (Array.init i (fun i -> good_id ("a"^string_of_int i))) in
                    true, (X.StType(X.TyDef(false,[tuple_ty.(i),vars,[tuple_con.(i), List.map (fun x -> TyVar x) vars]])),
                    dummy_loc))
  in
  List.fold_left (fun (env,genenv,ornenv,outdefs) (_,(X.StType(X.TyDef(is_global,[def])),_)) ->
    let env, genenv, outdef = OI.elaborate_tydef env genenv def in
    let (i,_,_) = def in
    let ornenv = Orn_codegen.elaborate_identity dummy_loc env genenv ornenv [i] in
    env,
    genenv,
    ornenv,
    outdefs@[false,(X.StType(X.TyDef(is_global,[def])), dummy_loc)]@[false,(X.StType(X.TyDef(false,[outdef])), dummy_loc)]
  ) (initial_ml_env, initial_gen_env, Orn.initial_orn_env, initial_defs) initial_defs


let declare_type env tydef =
  let (tydef, scope) = Ml_inference.elaborate_tydefs env.scope tydef in
  let X.TyDef(_,defs) = tydef in
  let dataenv_entries = List.map (fun (tn, targs, conses) -> (tn, (targs, conses))) defs in
  let conenv_entries = List.concat (List.map (fun (tn, targs, conses) -> List.map (fun (c,args) -> (c, (tn,targs,args))) conses) defs) in
  {env with
    scope = scope;
    dataenv = dataenv_entries@env.dataenv;
    conenv = conenv_entries@env.conenv
  }, tydef


let ty_free_vars =
  let rec ftv = function
    | TyVar v -> IdSet.singleton v
    | TyArrow(t1,t2) -> IdSet.union (ftv t1) (ftv t2)
    | TyProduct(ts) | TyCon(_,ts) -> List.fold_left (fun s t -> IdSet.union s (ftv t)) IdSet.empty ts
  in
  ftv


let load_toplet (env, genenv, ornenv, indefs, outdefs, envstack) isrec bindings =
  debug "Typechecking %a@." (seq ", "Pp.user_id) (List.map (fun (x,_) -> x) bindings);
  let env, X.TopLet(isrec,bindings) = Ml_inference.elaborate_def env isrec bindings in
  let bindings = List.map (fun (x,sc,e) ->
    let e = Desugaring.do_all env e in
    debug "Desugared %a=@.%a@." Pp.id x Pp.xterm e;
    (x,sc,e)
  ) bindings in
  let toplet = X.TopLet(isrec, bindings) in
  let genenv, elabs, rn_rec (* list (x, osc, (tvs, reqs, e_gen)) *) = OI.elaborate_valdef genenv toplet in
  List.iter (fun (x,osc,_) ->
    say "generalized to [val %a : %a]@." Pp.id x Pp.osc osc;
  ) elabs;
  (* Now we have a bunch of things whose identity instanciations we need to compute *)
  (* Computing the identity instanciation _should_ be done through the normal instaciation mechanism:
     let's allow this !
     this requires separating the constraint-checking/unification thing and the instanciation thing,
     but this is good anyway, it will augment the trust in the identity instanciation *)
  say "Building the identity instanciations@.";
  (* To compute the identity instanciation, we... just do the obvious thing.
     There may be some difficulties about glueing the things back together *)
  (* We need to instanciate all recursive calls with the function itself.
     How do we find the right instance for mutually recursive calls ?
    => here we would want to propagate the instanciation constraint from the
       original term but we are missing the correct information.
    But we can solve the constraints considering that the functions are in the
     same recursive group *)
  let module U = Unifier.Make(Ml_inference.S) in
  let id_insts = try
    (* XXX Significant code duplication with do_lifting XXX *)
    let module O = Ml_inference.O in
    let module S = Ml_inference.S in
    let decode_variable x = X.OVar (U.id x) in
    let structure = function
      | S.TyCon(i, xs) -> X.OIdOrn(i, xs) (* Only id ornament appears. *)
      | S.TyArrow(t1, t2) -> X.OArrow(t1, t2)
    in
    let orn_decoder = U.new_acyclic_decoder decode_variable structure in
    let rec encode_oty map = function
      | X.OVar(x) -> IdMap.find x map
      | X.OArrow(t1, t2) -> U.fresh (Some(S.TyArrow(encode_oty map t1, encode_oty map t2))) 0
      | X.OBaseOrn(i,ts)
      | X.OIdOrn(i,ts) -> U.fresh (Some(S.TyCon(i, List.map (encode_oty map) ts))) 0
      | X.OProduct ts -> U.fresh (Some(S.TyCon( tuple_ty.(List.length ts), List.map (encode_oty map) ts ))) 0
    in
    let rec encode_ty map = function
      | TyVar x -> IdMap.find x map
      | TyArrow(t1, t2) -> U.fresh (Some(S.TyArrow(encode_ty map t1, encode_ty map t2))) 0
      | TyCon(i,ts) -> U.fresh (Some(S.TyCon(i, List.map (encode_ty map) ts))) 0
      | TyProduct ts -> U.fresh (Some(S.TyCon( tuple_ty.(List.length ts), List.map (encode_ty map) ts ))) 0
    in
    (* XXX end of duplication XXX *)
    let scvars = List.concat (List.map (fun (_,osc,_) -> osc.Orn.ornscheme_paravars) elabs) in
    let scorns = List.concat (List.map (fun (_,osc,_) -> osc.Orn.ornscheme_ornvars) elabs) in
    let varmap = List.fold_left (fun m x -> IdMap.add x (U.fresh None 0) m) IdMap.empty scvars in
    let varmap = List.fold_left (fun m (x,_,_) -> IdMap.add x (U.fresh None 0) m) varmap scorns in
    (* Instantiate all ornament variables with the identity ornament *)
    List.iter (fun (x,i,os) ->
      let i_skel = List.assoc i genenv.X.skel_map in
      let skel_desc = List.assoc i_skel genenv.X.skel_dataenv in
      let svarmap = List.fold_left (fun m x -> IdMap.add x (U.fresh None 0) m) IdMap.empty skel_desc.X.skel_id_args in
      U.unify (IdMap.find x varmap) (U.fresh (Some( S.TyCon(i, List.map (fun x -> IdMap.find x svarmap) skel_desc.X.skel_id_args) )) 0);
      List.iter2 (fun t o -> U.unify (encode_ty svarmap t) (encode_oty varmap o)) skel_desc.X.skel_id_inst os
    ) scorns;
    (* XXX XXX this is probably broken because we do not consider correctly the recursive constraints. *)
    (* We need to find the list of interesting variables we should quantify over.
       The simplest is to list the free variables in targs/orns *)
    List.map (fun (x,osc,_) ->
      let lifting_insts = List.map (fun x -> orn_decoder (IdMap.find x varmap)) osc.Orn.ornscheme_paravars in
      let lifting_orns = List.map (fun (x,_,_) -> orn_decoder (IdMap.find x varmap)) osc.Orn.ornscheme_ornvars in
      let lifting_targs =
        IdSet.elements (
          List.fold_left (fun s t -> IdSet.union s (X.oty_free_vars t))
            (List.fold_left (fun s t -> IdSet.union s (X.oty_free_vars t)) IdSet.empty lifting_insts)
            lifting_orns
        ) in
      (x,X.{ lifting_of = x; lifting_targs; lifting_insts; lifting_orns })
    ) elabs
    with
    | (U.Unify _ | U.Cycle _) when not !debug_mode -> Error.failwith "Internal error"
  in
  let ornenv =
    let generic_terms = List.map (fun (x,osc,(tvs,reqs,e_gen)) ->
      let gen_descr = Orn.{
        generic_tvars = tvs;
        generic_ornscheme = osc;
        generic_requests = reqs;
        generic_rec_renaming = rn_rec;
        generic_term = e_gen
      } in
      (x, gen_descr)
    ) elabs in
    let id_liftings = List.map (fun (x,id_inst) -> ((x,x), id_inst)) id_insts in
    let orn_scope = List.fold_left (fun orn_scope (x,_,_) -> Scope.(intro_named orn_scope NsVar (UserId x.id_name) x) ) ornenv.Orn.orn_scope elabs in
    Orn.{ornenv with
      orn_generic = generic_terms@ornenv.orn_generic;
      env_liftings = (if !bind_names then id_liftings else []) @ ornenv.env_liftings;
      orn_scope;
    }
  in
  env,
  genenv,
  ornenv,
  indefs @ [X.StVal(toplet), dummy_loc],
  (if !echo_decls then outdefs@[true, (X.StVal(toplet), dummy_loc)] else outdefs),
  envstack


(* How do I design this ?
   Unification engine: takes instanciation (partial if necessary),
   checks the constraints (and update about them if possible),
   is rollbackable,
   finalizes into usable code + instanciation information *)

module Lifter(O : sig val oenv : Ast.Explicit.orn_env end) = struct
  module U = Orn_unifier.Make(O)
  module S = U.S
  module O = struct
    let orn_to_ty =
      let rec cv = function
        | X.OVar i -> TyVar i
        | X.OArrow(t1, t2) -> TyArrow(cv t1, cv t2)
        | X.OBaseOrn(c, ts) -> TyCon(c, List.map cv ts)
        | _ -> assert false
      in
      cv

    let variable x =
      X.OVar x

    let structure = function
      | S.TyArrow(t1,t2) -> X.OArrow(t1,t2)
      | S.TyOrn(s, l) -> X.OBaseOrn(s, l)
  end

  let decode_variable x = O.variable (U.id x)
  let orn_decoder v = U.new_acyclic_decoder decode_variable O.structure v
  let new_decoder v = O.orn_to_ty (orn_decoder v)

  let rec encode_oty map = function
    | X.OVar(x) -> IdMap.find x map
    | X.OArrow(t1, t2) -> U.fresh (Some(S.TyArrow(encode_oty map t1, encode_oty map t2)))
    | X.OBaseOrn(i,ts)
    | X.OIdOrn(i,ts) -> U.fresh (Some(S.TyOrn(i, List.map (encode_oty map) ts)))
    | X.OProduct ts -> assert false

  let rec encode_ty map = function
    | TyVar x -> (try IdMap.find x map with Not_found -> Error.failwith "Type variable not found: %a" Pp.id x)
    | TyArrow(t1, t2) -> U.fresh (Some(S.TyArrow(encode_ty map t1, encode_ty map t2)))
    | TyCon(i,ts) -> U.fresh (Some( S.TyOrn(i, List.map (encode_ty map) ts) ))
    | TyProduct ts -> U.fresh (Some( S.TyOrn(tuple_ty.(List.length ts), List.map (encode_ty map) ts) ))

  let intro_vars l = List.fold_left (fun m x -> IdMap.add x (U.fresh None) m) IdMap.empty l

  (* Local liftings are not generalized *)
  type local_lifting_descr = {
    loclift_of : id;
    loclift_insts : U.variable list;
    loclift_orns : U.variable list
  }

  type maybe_local_lifting =
    | GlobalLifting of Orn.lifting_descr
    | LocalLifting of local_lifting_descr

  exception Blocked
end

exception Rollback
exception Update



(* Lifting! We use the unifier directly, it is more convenient than calling the solver repeatedly.
   We cannot just work with the solver of ml_inference : we need to keep track of the skeletons
   to track all the constraints (so we are efficient + sound) *)
(* Let's recall the encoding:
    everything is encoded normally
    except ornament types that are encoded as TyCon(i_skel,(args of skeleton), TyCon(o, ts))
   and 'any ornament of bla' is encoded as TyCon(i_skel, (args of skeleton), <var>)
*)

(*
rigid lifting mode (single recursion only):
  self calls stay self calls
*)
let liftcount = ref 1
let do_lifting loc (env,genv,oenv,indefs,outdefs,envstack) (ldefs : lift_def list) =
  let module Lifter = Lifter(struct let oenv = oenv end) in
  let open Lifter in
  let open Orn.Implicit in
  let open Scope in
  let module E = Eml.ExplicitGml in
  (* We use only one constraint solver for all definitions.
     They are all (temporarily) monomorphic, we generalize the whole group
     at once afterwards.
     This removes problems with polymorphic recursion.
     Auto-{lifting,unfolding} might end up awkward
      ---> match only on the ornament names (ie. do not autolift twice with the same ornament names)
  *)
  (* How it works:
     - we have a list of things to instantiate that are named
     - we add them to the environment, making them available for var instanciation (auto or not)
     - we also make the special recursive instanciation available
       (XXX these should be marked in a special way)
     - we walk the constraints
     - when no var intanciation is available, we create a new auto-lifting
       * add it to the environment
       * add the constraint that it correctly instantiates
     - this might not terminate, and that's alright
  *)
  (* Type variables are shared between all simultaneously-defined things if not eplicitly quantified *)
  let ospec_vars = Hashtbl.create 5 in
  let rec encode_ospec = function
    | TyVar(UserId("_")) -> U.fresh None
    | TyVar(x) ->
     (* We do not support explicit quantification, so the variable could end up specialized...
        Although we force all 'a to be the same
     *)
       begin try Hashtbl.find ospec_vars x
         with Not_found ->
           let v = U.fresh None in
           Hashtbl.add ospec_vars x v;
           v
       end
    | TyArrow(t1, t2) -> U.fresh (Some(S.TyArrow( encode_ospec t1, encode_ospec t2 )))
    | TyCon(o, [TyVar(UserId("_"))]) ->
     (* The user can write `_ either ' instead of `(_,_) either' and we take care of finding the arity *)
       let o = resolve oenv.Orn.orn_scope NsType o in
       let (o_def,_,_,_) = List.assoc o oenv.X.env_ornaments in
       U.fresh (Some( S.TyOrn(o, List.map (fun _ -> U.fresh None) o_def.X.orn_args) ))
    | TyProduct(ts) ->
       U.fresh (Some( S.TyOrn(tuple_ty.(List.length ts), List.map encode_ospec ts) ))
    | TyCon(o, ts) ->
       let o = resolve oenv.Orn.orn_scope NsType o in
       U.fresh (Some( S.TyOrn(o, List.map encode_ospec ts) ))
  in
  (* For user-requested liftings: lookup from_name, add to_name to scope, unify the things, add the patches
     For other liftings: not necessary.
     We start with only allowing user-defined lifting *)
  let init_new_lifting env (* here: env.scope *) (l : lift_def) =
    let from_name = resolve env.scope NsVar l.lift_from in
    let to_name, scope = intro env.scope NsVar l.lift_to in
    let genspec = List.assoc from_name oenv.Orn.orn_generic in
    let Orn.{
      generic_tvars = tvs;
      generic_ornscheme = {
        ornscheme_paravars = sc_tvars;
        ornscheme_ornvars = sc_ovars;
        ornscheme_ornty = to_ot};
      generic_requests = inst_reqs;
      generic_term = from_gen;
      generic_rec_renaming = rec_rn;
    } = genspec in
    (* Let's go! We are interested in producing:
       - a concrete ornament for all the instanciation requests
       - exact values for the variables of the scheme so we can put the signature back somewhere *)
    let tvar_map = intro_vars tvs in
    List.iter (fun (x,i,ts) ->
      let i_skel = List.assoc i genv.X.skel_map in
      U.unify (IdMap.find x tvar_map) (U.fresh_skel (U.Skeleton(i_skel, List.map (encode_oty tvar_map) ts)))
    ) sc_ovars;
    (* We can use the _types_ to guess at some part of the mapping *)
    let to_ov = encode_oty tvar_map to_ot in
    let ospec_v = encode_ospec (match l.lift_at with Some(t) -> t | None -> TyVar(UserId("_"))) in
    begin
      try
        U.unify to_ov ospec_v
      with
        U.Unify(v1,v2) when not !debug_mode->
          Error.error loc ("Ornament specification does not match the shape of the original type.\n"^^
                              "Could not unify %a and %a.")
            Pp.ty(*_nice*) (new_decoder v1) Pp.ty(*_nice*) (new_decoder v2)
    end;
    {env with scope = scope}, (from_name, to_name, tvar_map, to_ov, genspec)
  in
  let liftings_scope = oenv.X.orn_scope in
  (* We will have to do something with special liftings *)
  let env, liftings_scope, bindings =
    List.fold_left (fun (env, liftings_scope, bindings) l ->
      (* XXX this is duplicated *)
      let env, ((_,to_name,_,_,_) as extra) = init_new_lifting env l in
      env, Scope.(intro_named liftings_scope NsVar l.lift_to to_name),(to_name,(l, extra))::bindings
    ) (env, liftings_scope, []) ldefs
  in
  let bindings = ref bindings in
  (* Make available the locally defined liftings *)
  let local_liftings = List.map (fun (key,ldesc) -> (key, GlobalLifting ldesc)) oenv.Orn.env_liftings in
  let local_liftings = List.fold_left (fun local_liftings (_,(_,(from_name,to_name,tvar_map,_,genspec))) ->
    (* XXX this is duplicated *)
    let map_vars l = List.map (fun x ->
        IdMap.find x tvar_map
    ) l
    in
    let ldesc = {
      loclift_of = from_name;
      loclift_insts = map_vars Orn.(genspec.generic_ornscheme.ornscheme_paravars);
      loclift_orns = map_vars (List.map (fun (x,_,_) -> x) Orn.(genspec.generic_ornscheme.ornscheme_ornvars));
    }
    in
    ((from_name, to_name), LocalLifting ldesc)::local_liftings
  ) local_liftings !bindings in
  let local_liftings = ref local_liftings in
  (* The liftings used are not recoverable from the unifier state,
     we store them in a hashtable *)
  let liftings = Hashtbl.create 5 in
  let apply_lifting_transactionally t lifting_what (n,x_lift,x_base,insts,orns) loc y_lift =
    let _, (_,_,tvar_map,_,genspec) = List.assoc lifting_what !bindings in
    let (y_a, y_b) = y_lift in
    let rec_rn = genspec.Orn.generic_rec_renaming in
    let (x_base, inst_vs, orn_vs) =
      match List.find_opt (fun (x_rec,x_unrec) -> x_base = x_unrec) rec_rn with
      | None ->
         (x_base, List.map (encode_oty tvar_map) insts, List.map (encode_oty tvar_map) orns)
      | Some (x_base, x_unrec) ->
         (* The y_desc is described by its insts/orns, but the request only gives the whole type.
            So what we do is we create fresh variables for the insts/orns as in the gendesc,
            encode the whole ornament type given in the request and the one in the gendesc;
            then we can work as before but unifying the variables *)
         debug "took the special path %a %a@." Pp.id x_base Pp.id x_unrec;
         let Orn.{generic_ornscheme = x_osc} = List.assoc x_base oenv.Orn.orn_generic in
         let ovar_map = List.fold_left (fun ovar_map x ->
           IdMap.add  x (U.fresh None) ovar_map
         )
           IdMap.empty x_osc.X.ornscheme_paravars
         in
         let ovar_map = List.fold_left (fun ovar_map (x,orn_i,orn_args) ->
           IdMap.add x (U.fresh None) ovar_map
         )
           ovar_map x_osc.X.ornscheme_ornvars
         in
         List.iter (fun (x,i,orn_args) ->
           let i_skel = List.assoc i genv.X.skel_map in
           U.unify_transactionally t (IdMap.find x ovar_map)
             (U.fresh_skel (U.Skeleton(i_skel, List.map (encode_oty ovar_map) orn_args)))
         ) x_osc.X.ornscheme_ornvars;
         let oty_actual_v = encode_oty ovar_map x_osc.X.ornscheme_ornty in
         let [oty_requested] = insts in
         let oty_requested_v = encode_oty tvar_map oty_requested in
         U.unify_transactionally t oty_actual_v oty_requested_v;
         (x_base,
          List.map (fun x -> IdMap.find x ovar_map) x_osc.X.ornscheme_paravars,
          List.map (fun (x,_,_) -> IdMap.find x ovar_map) x_osc.X.ornscheme_ornvars)
    in
    let y_desc =
      try
        List.assoc y_lift !local_liftings
      with Not_found ->
        Error.error loc "`%a' is not a lifting of `%a'" Pp.id y_b Pp.id y_a in
         (* "Localize" the description, instantiating/encoding if needed *)
    let y_desc = match y_desc with
      | LocalLifting ldesc -> ldesc
      | GlobalLifting y_desc ->
         let varmap = List.fold_left (fun m x -> IdMap.add x (U.fresh None) m)
           IdMap.empty y_desc.X.lifting_targs in
         {
           loclift_of = y_desc.X.lifting_of;
           loclift_insts = List.map (encode_oty varmap) y_desc.X.lifting_insts;
           loclift_orns = List.map (encode_oty varmap) y_desc.X.lifting_orns;
         }
    in
    if y_desc.loclift_of <> x_base then raise S.Iter2;
    if not !enable_unfolding then begin
      let _,(from_name,to_name,_,_,_) = List.assoc lifting_what !bindings in
      if from_name = x_base && to_name <> y_b then raise S.Iter2;
    end;
    List.iter2 (U.unify_transactionally t) y_desc.loclift_insts inst_vs;
    List.iter2 (U.unify_transactionally t) y_desc.loclift_orns orn_vs
  in
  let try_apply_lifting keep_applied lifting_what (n,x_lift,x_base,insts,orns) loc y_lift =
    try
      TRef.tentatively (fun t ->
        apply_lifting_transactionally t lifting_what (n,x_lift,x_base,insts,orns) loc y_lift;
        if keep_applied
        then (debug "<- ok@."; Hashtbl.add liftings (lifting_what, n) y_lift; true)
        else raise Rollback
      )
    with
    | Rollback -> true
    | S.Iter2 | U.Unify _ -> false
  in
  (* We might need a way to disable unique liftings *)
  let seen_new_bindings = ref false in
  let rec find_unique_liftings lifting_what =
    let _, (_,_,_,_,genspec) = List.assoc lifting_what !bindings in
      List.iter (function
      | X.InstantiateEnv(x_loc,n,x_lift,x_base,insts,orns) when not (Hashtbl.mem liftings (lifting_what,n))->
         let request = (n,x_lift,x_base,insts,orns) in
         begin match List.filter (fun (y_lift,y_desc) ->
           try_apply_lifting false lifting_what  request dummy_loc y_lift
         ) !local_liftings with
         | [((x_base,y_lift),_)] ->
            if try_apply_lifting true lifting_what request dummy_loc (x_base,y_lift)
            then (say "[lifting #%d of %a] instanciated with unique lifting [%a]" n Pp.id_nice x_base Pp.id_nice y_lift; raise Update)
            else failwith "Internal error"
         | [] -> begin
            try
            debug "(no lifting available for [lifting #%d of %a])" n Pp.id_nice x_base;
           if !enable_autolifting then begin
             let y_base =
                match List.find (fun (_,x_unrec) -> x_base = x_unrec) genspec.X.generic_rec_renaming with
                | (x_rec,_) -> x_rec
                | exception Not_found -> x_base
             in
             (* We duplicate the check otherwise we will create infinitely many liftings that will be blocked
                from actually instanciating. XXX rework this flow  *)
             if not !enable_unfolding then begin
               let _,(from_name,to_name,_,_,_) = List.assoc lifting_what !bindings in
               if from_name = y_base then begin
                 debug "(autolifting blocked by no-unfolding)@.";
                 raise Blocked
               end
             end;
             (* Create a new lifting and then instanciate with it *)
             let y_lift = id (y_base.id_name ^ "_auto" ^string_of_int !liftcount) in
             incr liftcount;
             let l = {
               lift_from = FromId y_base;
               lift_to = FromId y_lift;
               lift_at = None;
               lift_with = [];
               lift_loc = x_loc
             } in
             (* XXX this is duplication *)
             let _, ((from_name,to_name,tvar_map,_,genspec) as extra) = init_new_lifting env l in
             bindings := (to_name, (l, extra)) :: !bindings;
             let map_vars l = List.map (fun x ->
               IdMap.find x tvar_map
             ) l
             in
             let ldesc = {
               loclift_of = from_name;
               loclift_insts = map_vars Orn.(genspec.generic_ornscheme.ornscheme_paravars);
               loclift_orns = map_vars (List.map (fun (x,_,_) -> x) Orn.(genspec.generic_ornscheme.ornscheme_ornvars));
             }
             in
             local_liftings := ((from_name, to_name), LocalLifting ldesc)::!local_liftings;
             seen_new_bindings := false;
             raise Update
           end
            with Blocked -> () end
         | l -> debug "(multiple liftings available for [lifting #%d of %a]: %a" n Pp.id_nice x_base (seq ", " Pp.id_nice) (List.map (fun ((_,x),_) -> x) l)
         end
      | _ -> ()
      ) genspec.generic_requests
  in
  let rec find_unique_liftings_all () =
    try
      seen_new_bindings := true;
      List.iter (fun (_,(_, (_, lifting_what,_,_,_))) -> find_unique_liftings lifting_what) !bindings
    with Update -> find_unique_liftings_all ()
  in
  find_unique_liftings_all ();
  let match_addr a patch_id = match a with
    | PatchFor(l) -> List.mem patch_id l
    | DefaultPatch -> true
  in
  List.iter (fun (lifting_what,(l,(_,_,tvar_map,_,gdesc))) ->
    List.iter (function
    | PatchOrnament(a, os), loc ->
       let is_default = a = DefaultPatch in
       List.iter (function
       | Orn.InstantiateOrnVar(_,n,x,i,_,_,_,_) when match_addr a n ->
          if not
            (List.exists (fun o ->
              let o = match o with
                | PatchSpecial(UserId "id") -> FromId i
                | PatchSpecial(UserId "exhaustive") -> Error.error loc "Ornament list is not exhaustive for [ornament #%d] of [%a]" n Pp.id_nice i
                | PatchIdent(o) -> o
                | PatchSpecial s -> Error.error loc "Unsupported directive %a" Pp.user_id s
              in
              let o = resolve_l loc oenv.X.orn_scope NsType o in
              let odef, _, _, _ = List.assoc o oenv.X.env_ornaments in
              let args = List.map (fun _ -> U.fresh None) odef.X.orn_args in
              try U.unify (IdMap.find x tvar_map) (U.fresh (Some( S.TyOrn(o,args) )));
                  say "[ornament #%d of %a] has been instantiated with [%a]" n Pp.id_nice i Pp.id_nice o;
                  true
              with
              | U.Unify(v1,v2) -> say "[ornament #%d of %a] could not be instantiated with [%a]" n Pp.id_nice i Pp.id_nice o;
                false
            ) os) && not is_default
          then Error.error loc "Could not instantiate [ornament #%d of %a] with any of the provided ornaments" n Pp.id_nice i
       | _ -> ()
       ) gdesc.Orn.generic_requests;
       find_unique_liftings_all ()
    | PatchLifting(a, ls), loc ->
       let is_default = a = DefaultPatch in
       List.iter (function
       | Orn.InstantiateEnv(_,n,x_lift,x_base,insts,orns) when match_addr a n ->
          if not
            (List.exists (fun y_lift ->
              let y_lift = match y_lift with
                | PatchSpecial(UserId "id") -> FromId x_base
                | PatchSpecial(UserId "exhaustive") -> Error.error loc "Lifting list is not exhaustive for [lifting #%d of %a]" n Pp.id_nice x_base
                | PatchIdent(y_lift) -> y_lift
                | PatchSpecial s -> Error.error loc "Unsupported directive %a" Pp.user_id s
              in
              let y_lift = resolve liftings_scope NsVar y_lift in
              let y_base =
                match List.find (fun (_,x_unrec) -> x_base = x_unrec) gdesc.X.generic_rec_renaming with
                | (x_rec,_) -> x_rec
                | exception Not_found -> x_base
              in
              if try_apply_lifting true lifting_what (n,x_lift,x_base,insts,orns) loc (y_base,y_lift)
              then (say "[lifting #%d] has been instantiated with [%a]" n Pp.id_nice y_lift; true)
              else (say "could not instantiate [lifting #%d of %a] with [%a]" n Pp.id_nice x_base Pp.id_nice y_lift; false)
             ) ls) && not is_default
          then Error.error loc "Could not instantiate [lifting #%d of %a] with any of the provided liftings" n Pp.id_nice x_base;
       | _ -> ()) gdesc.Orn.generic_requests;
       find_unique_liftings_all ()
    | PatchHole _,  loc -> () (* We cannot learn anything from this *)
    ) l.lift_with
  ) !bindings;
  while not !seen_new_bindings do find_unique_liftings_all () done;
  let bindings = List.rev !bindings in
  (******************************************************)
  (**** Everything is determined, we can instantiate ****)
  (******************************************************)
  let new_bindings = List.map (fun (lifting_what, (l, (from_name, to_name, tvar_map, to_ov, genspec))) ->
    let gty = E.TVar(id "##UNTYPED##") in
    let substs_te, substs_ty =
      List.split(
        List.map
          (function
          | Orn.InstantiateEnv(_,n,p,x_base,_,_) ->
             IdHashtbl.add Pretty.holemap p ("(lifting #"^string_of_int n^" of "^x_base.id_name^")");
            begin match Hashtbl.find liftings (lifting_what,n) with
            | _, y_lift -> [(p, (E.Var y_lift, gty, dummy_loc))], []
            | exception Not_found ->
               say "no lifting given/inferred for [lifting #%d of %a]" n Pp.id_nice x_base;
              [], []
            end
          | Orn.InstantiateOrnVar(_,n,ovar,c,_,xext,xcon,xproj) ->
             IdHashtbl.add Pretty.holemap xcon ("(orn-cons #"^string_of_int n^")");
            IdHashtbl.add Pretty.holemap xproj ("(orn-match #"^string_of_int n^")");
            begin match new_decoder (IdMap.find ovar tvar_map) with
            | TyCon(orn_name, _) ->
               let (orn_descr, ext, con, proj) = List.assoc orn_name oenv.Orn.env_ornaments in
               say "using ornament %a for [ornament #%d of %a]" Pp.id_nice orn_name n Pp.id_nice c;
               [(xcon, con); (xproj, proj)], [(xext,ext)]
            | _ -> say "no ornament given/inferred for [ornament #%d of %a]" n Pp.id_nice c; [], []
            end
          | Orn.IdentityOrnament(c,_,xext,xcon,xproj) ->
             let (_, ext, con, proj) = List.assoc c oenv.Orn.env_ornaments in
             [(xcon, con); (xproj, proj)], [(xext,ext)]
          | Orn.InstantiatePatch(vars,_,n,x,_) ->
             IdHashtbl.add Pretty.holemap x ("#"^string_of_int n);
            begin match search  l.lift_with (function PatchHole(a,e), _ when match_addr a n -> Some(e) | _ -> None) with
            | e ->
               let e = Ml_inference.fix_contuple env e in
               let e = Orn_codegen.name_patch env.scope vars e in
               let e = Desugaring.untyped_all_patch env e in
               let e = Eml_compute.ExplicitToGml.ml_to_gml e in
               [(x,e)], []
            | exception Not_found ->
               say "no patch given for [#%d]" n; [], []
            end
          ) genspec.Orn.generic_requests)
    in
    let te_env, ty_env = List.concat substs_te, List.concat substs_ty in
    debug "Substs: (%a) and (%a)@." (seq "," Pp.id) (List.map fst te_env) (seq "," Pp.id) (List.map fst ty_env);
    (* We apply this substitution and evaluate the eml term *)
    let reduced = Eml_compute.GmlReduction.reduce_te ty_env te_env genspec.Orn.generic_term in
    debug "%a@." Pp.eml_expr reduced;
    let t = Eml_compute.EmlToImplicit.cv reduced in
    (l,from_name,to_name,tvar_map,to_ov,genspec,t)
  ) bindings in
  let new_bindings = if !enable_inlining then Eml_untyped.inline_autodefs new_bindings else new_bindings in
  let new_bindings = List.map (fun (l,from_name,to_name,tvar_map,to_ov,genspec,t) ->
    let t = Desugaring.ANFTransform.transform t in
    let t = Eml_untyped.elementary_simplification t in
    debug "Generated eml term:\n%a@." Pp.term t;
    let t = Eml_untyped.match_reduction [] t in
    let t = Eml_untyped.elementary_simplification t in
    debug"Match-reduced to:\n%a@." Pp.term t;
    let t = Eml_untyped.resugar t in
    debug "Resugared to:\n%a@." Pp.term t;
    (l,from_name,to_name,tvar_map,to_ov,genspec,t)
  ) new_bindings in
  match
    let new_bindings = List.map (fun (l,fn,tn,tm,tv,gs,t) -> (l,fn,tn,tm,tv,gs,Eml_untyped.eliminate_labels t)) new_bindings in
    let bs = List.map (fun (_,_,tn,_,_,_,t) -> (tn, sc, t)) new_bindings in
    let toplet = X.TopLet(true, bs) in
    let toplet = Eml_untyped.unrecursify_toplet toplet in
    let isrec, def_untyped = Ast.untype toplet in
    toplet, load_toplet (env,genv,oenv,indefs,outdefs,envstack) isrec def_untyped, new_bindings
  with
  | exception Eml_untyped.HasHoles ->
     List.iter (fun (l,_,_,_,_,_,_) ->
       let UserId n = l.lift_to in
       if n.[0] <> '_' then
         Error.warning loc "Lifted term `%s' still has holes. Name it `_%s' to silence this warning" n n
     ) new_bindings;
    let bs = List.map (fun (_,_,tn,_,_,_,t) -> (tn, sc, t)) new_bindings in
    let toplet = X.TopLet(true, bs) in
    let toplet = Eml_untyped.unrecursify_toplet toplet in
    env, genv, oenv, indefs,
    outdefs@(if !bind_names then [!output_things, (X.StVal(toplet), dummy_loc)] else []),
    envstack
  | exception e when not !debug_mode -> raise (Patch_error e)
  | toplet, (env,genv,oenv,indefs,outdefs,envstack), bindings  ->
     (* We now only have to determine the instanciations *)
     let liftings = List.map (fun (_,from_name,to_name,tvar_map,to_ov,genspec,t) ->
       let Scheme(spec_tvs,spec_ty) = List.assoc to_name env.varenv in
       let varmap = List.fold_left (fun m x -> IdMap.add x (U.fresh None) m) IdMap.empty spec_tvs in
       let spec_v = encode_ty varmap spec_ty in
       let rec upr v =
         match U.structure v with
         | None -> v
         | Some( S.TyOrn(o, ts) ) ->
              (* We need to encode the right-part of this ornament, and be clever about it ... *)
            let odesc, _, _, _ = List.assoc o oenv.X.env_ornaments in
            let varmap = List.fold_left2 (fun m x v -> IdMap.add x (upr v) m) IdMap.empty odesc.X.orn_args ts in
            encode_ty varmap odesc.X.orn_to
         | Some(t) -> U.fresh (Some( S.map upr t ))
       in
       let to_v = upr to_ov in
       U.unify to_v spec_v;
       let lifting_insts = List.map (fun x -> orn_decoder (IdMap.find x tvar_map)) Orn.(genspec.generic_ornscheme.ornscheme_paravars) in
       let lifting_orns = List.map (fun (x,_,_) -> orn_decoder (IdMap.find x tvar_map)) Orn.(genspec.generic_ornscheme.ornscheme_ornvars) in
       let lifting_targs =
         IdSet.elements (List.fold_left (fun s o -> IdSet.union s (X.oty_free_vars o)) IdSet.empty (lifting_insts @ lifting_orns)) in
       let lift_descr = X.{
         lifting_of = from_name;
         lifting_targs;
         lifting_insts;
         lifting_orns;
       } in
       ((from_name,to_name),lift_descr)
     ) bindings
     in
     env,
     genv,
     Orn.{oenv with env_liftings = liftings@oenv.env_liftings},
     indefs,
     outdefs@(if !bind_names then [!output_things, (X.StVal(toplet), dummy_loc)] else []),
     envstack

(* We output the final term *)

    
(* Useful for debugging... *)
let dump_liftings oenv =
  debug "Currently defined liftings:";
  List.iter (fun ((x,y),l) ->
    debug "%a [%a], lifting of %a with instances [%a] / [%a]" Pp.id y Pp.(seq ", " tyvar) l.X.lifting_targs Pp.id x
      Pp.(seq ", " oty) l.X.lifting_insts Pp.(seq ", " oty) l.X.lifting_orns
  ) oenv.X.env_liftings


let process_ml_file dotml_path dotml_ast (env,genenv,ornenv,output) = 
  let env,genenv,ornenv, input, output, _ = List.fold_left (fun ((env, genenv, ornenv, indefs, outdefs, envstack) as envs) ->
    function phrase, phrase_loc ->
      dump_liftings ornenv;
      try
        match phrase, phrase_loc with
        | StPragma([]), loc -> Error.error loc "empty pragma"
        | StPragma([UserId "push_env"]), loc ->
           (env,genenv,ornenv,indefs,outdefs, (env,genenv,ornenv)::envstack)
        | StPragma([UserId "pop_env"]), loc ->
           begin match envstack with
           | (env,genenv,ornenv)::envstack -> (env, genenv, ornenv, indefs, outdefs, envstack)
           | [] -> Error.error loc "environment stack empty"
           end
        | StPragma(head::tail), loc ->
           begin
             match head with
             | UserId "mode" ->
                begin match tail with
                | [UserId flag] -> process_flag flag
                | _ -> Error.error loc "Unknown flag set"
                end
             | _ -> Error.error loc "Unknown pragma"
           end;
          envs
        | StVal(isrec, bindings), _ ->
           let bindings = List.map (fun (x,e) -> (x,fix_contuple env e)) bindings in
           load_toplet envs isrec bindings
        | StAbstractVal(x,t), _ ->
           let env, x, sc = elaborate_abstract_def env x t in
           let Scheme(scvars, ty) = sc in
           debug "val %a : %a@." Pp.id x Pp.tyscheme sc;
           let genenv = OI.elaborate_abstract_def genenv x sc in
           let lifting_descr = X.{
             lifting_of = x;
             lifting_targs = scvars;
             lifting_insts = List.map (fun x -> OVar(x)) scvars;
             lifting_orns = []
           } in
           (* We should save this as an "actually ornamented thing". But we don't have this yet *)
           env,
           genenv,
           Orn.{ornenv with env_liftings = ((x,x),lifting_descr)::ornenv.env_liftings},
           indefs@[X.StAbstractVal(x, sc), dummy_loc],
           outdefs@(if !bind_names then [!output_things,(X.StAbstractVal(x, sc), dummy_loc)] else []),
           envstack
        | StType(tydefs), loc ->
           (*                     let TyDef(x_user,_,_) = tydef in *)
           let env, def = declare_type env tydefs in
           let env, genenv, outdef = OI.elaborate_tydefs env genenv def in
           let X.TyDef(_,defs) = def in
           let is = List.map (fun (i,_,_) -> i) defs in
           let ornenv = Orn_codegen.elaborate_identity loc env genenv ornenv is in
           env,
           genenv,
           ornenv,
           indefs@[X.StType(def), dummy_loc; X.StType(outdef), dummy_loc],
           outdefs
           @(if !noisy_output || !bind_names then [!output_things && !bind_names,(X.StType(def), dummy_loc)] else [])
           @(if !noisy_output || true then [true,(X.StType(outdef), dummy_loc)] else []),
           envstack
        | StOrnDef(odef), loc ->
           let ornenv = Orn_codegen.elaborate_orndef loc env genenv ornenv odef in
           env, genenv, ornenv, indefs, outdefs, envstack
        | StLiftDef(ldefs), loc -> do_lifting loc envs ldefs
        | StCommented(_), _ -> env, genenv, ornenv, indefs, outdefs, envstack
      with
      | Wrong_arity loc when not !debug_mode -> error loc "Wrong arity"
      | Failure s when not !debug_mode -> error phrase_loc "Internal error: %s" s
      | Patch_error e when not !debug_mode -> print_error phrase_loc "Type error in patch caused the following exception"; raise e
      | User_error e when not !debug_mode -> error phrase_loc "%s" e
      | Error.Error _ as e when not !debug_mode -> raise e
      | e when not !debug_mode -> print_error phrase_loc "Internal error"; raise e
  ) (env, genenv, ornenv, [],output, []) dotml_ast
  in
  debug "Load ML file successful [%s]...@." dotml_path;
  (env, genenv, ornenv,output)

let load_ml_file dotml_path state =
  let dotml = open_in dotml_path in
  let dotml_lexbuf = Lexing.from_channel dotml in
  Util.curfile := dotml_path;
  let dotml_ast =
    try
      Parser.toplevel (Lexer.lex_with_errors dotml_path) dotml_lexbuf
    with
    | Parser.Error -> Error.(error (Loc.lex_loc dotml_path dotml_lexbuf) "Syntax error")
  in
  process_ml_file dotml_path dotml_ast state


(* We implicitely add prefixes wherever needed *)
let load_refactoring_spec rspec_path =
  let rspec = open_in rspec_path in
  let rspec_lexbuf = Lexing.from_channel rspec in
  Util.curfile := rspec_path;
  let rspec_ast =
    try
      Parser.toplevel (Lexer.lex_with_errors rspec_path) rspec_lexbuf
    with
    | Parser.Error -> Error.(error (Loc.lex_loc rspec_path rspec_lexbuf) "Syntax error")
  in
  let ornaments, types = List.fold_left (fun (ornaments, types) -> function
    | StType([tydef]), loc ->
       (ornaments, (tydef,loc)::types)
    | StType(_), loc ->
       Error.error loc "Type definitions must be written non-recursively (even if they are recursive)"
    | StOrnDef([odef]), loc ->
       ((odef,loc)::ornaments, types)
    | StOrnDef(_), loc ->
       Error.error loc "Ornament definitions must be written non-recursively (even if they are recursive)"
    | _, loc -> Error.error loc "Unexpected statement in refactoring spec"
  ) ([],[]) rspec_ast in
  (ornaments, types)

(* Process a file, generating a corresponding refactoring ast.
   Refactoring process:
    - load a refactoring spec
    - read the input and generate a refactoring file
    - read the input again
    - read the refactoring file
*)
let add_prefix p (UserId s) =
  UserId (p ^ s)

let gen_refactoring_ast (spec_orns, spec_types) dotml_path =
  let dotml = open_in dotml_path in
  let dotml_lexbuf = Lexing.from_channel dotml in
  Util.curfile := dotml_path;
  let dotml_ast =
    try
      Parser.toplevel (Lexer.lex_with_errors dotml_path) dotml_lexbuf
    with
    | Parser.Error -> Error.(error (Loc.lex_loc dotml_path dotml_lexbuf) "Syntax error")
  in
  let subst_types prefix =
    let rec st = function
      | TyCon(c, ts) -> TyCon(add_prefix prefix c, List.map st ts)
      | TyArrow(t1,t2) -> TyArrow(st t1, st t2)
      | TyProduct(ts) -> TyProduct(List.map st ts)
      | TyVar v -> TyVar v
    in
    st
  in
  let _PArg = function
    | [] -> []
    | l -> [PTuple(l), dummy_loc]
  in
 let _, output = List.fold_left (fun (orn_names,output) -> function
(*    | StVal(x,isrec,e), loc ->
       let ldef = {
         lift_from = x;
         lift_to = x;
         lift_loc = loc;
         lift_at = None;
         lift_with = [(PatchOrnament(DefaultPatch, List.map (fun x -> PatchIdent x) orn_names @ [PatchSpecial (UserId "id")]), dummy_loc)];
       } in
       orn_names, output @ [(StLiftDef ldef, loc)]
*)
    |StVal _, loc -> assert false
    | StType(tydefs), loc ->
    (* Generate the new definitions + an ornament *)
       let newdefs, neworns = List.split (
         List.map (fun (TyDef(tyname, tyargs, tyconses)) ->
           let newt = add_prefix "out_" tyname in
           let newo = add_prefix "orn_" tyname in
           match List.find (fun (od,_) -> od.orn_name = newo) spec_orns with
           | odef, loc -> (* Find the corresponding type *)
              (* Check source, arity, destination *)
              let args_ty = List.map (fun x -> TyVar x) odef.orn_args in
              if odef.orn_from <> TyCon(tyname, args_ty)
              then Error.error loc "The source of %a should be %a" Pp.user_ty (TyCon(newo, args_ty))
                Pp.user_ty (TyCon(tyname, args_ty));
              if odef.orn_to <> TyCon(newt, args_ty)
              then Error.error loc "The destination of %a should be %a" Pp.user_ty (TyCon(newo, args_ty))
                Pp.user_ty (TyCon(newt, args_ty));
              let tydef, _ =
                try List.find (function TyDef(tn,_,_), _ -> tn = newt) spec_types
                with Not_found -> Error.error loc "No definition of %a" Pp.user_id newt
              in
              (tydef, odef)
           | exception Not_found ->
              begin match List.find (function TyDef(tn,_,_),_ -> tn = newt) spec_types with
              | _, loc -> Error.error loc "This type definition cannot be used because there is no associated ornament"
              | exception Not_found -> ()
              end;
             (* copy the definition *)
             let tydef = TyDef(newt, tyargs, List.map (fun (c, ts) -> (add_prefix "Out_" c, List.map (subst_types "out_") ts)) tyconses) in
             let odef = {
               orn_name = newo;
               orn_args = tyargs;
               orn_from = TyCon(tyname, List.map (fun x -> TyVar x) tyargs);
               orn_to = TyCon(newt, List.map (fun x -> TyVar x) tyargs);
               orn_clauses = List.map (fun (c,ts) ->
                 let vars = List.mapi (fun i t -> (UserId ("x"^string_of_int i), subst_types "orn_" t)) ts in
                 ((PCon(c,_PArg (List.map (fun (v,_) -> PVar v, loc) vars)),loc),
                   vars,
                  (PCon(add_prefix "Out_" c, _PArg (List.map (fun (v,_) -> PVar v, loc) vars)) ,loc))) tyconses
               }
             in
             (tydef, odef)
         ) tydefs
       ) in
       let orn_names = List.map (fun odef -> odef.orn_name) neworns @ orn_names in
       orn_names, output @ [(StType newdefs, loc); (StOrnDef neworns, loc)]
    | (StOrnDef _ | StLiftDef _ | StCommented _ | StAbstractVal _ | StPragma _), loc ->
       Error.error loc "Unsupported construction in refactored file"
  ) ([],[]) dotml_ast in
  output

let print_output output =
  let output = PrettyNames.run_renamer (fun s ->
                                        ignore PrettyNames.(intro s NsCons sum_l);
                                        ignore PrettyNames.(intro s NsCons sum_r);
  List.fold_right (fun (is_printed,decl) k scope ->
    match decl with
      | X.StVal(v), loc -> Pretty.pretty_top is_printed scope loc v k
      | X.StType(t), loc ->
         Pretty.pretty_tydef is_printed scope t loc k
      | X.StAbstractVal(x, t), loc -> Pretty.pretty_absdef is_printed scope loc x t k
      | X.StCommented(v), loc -> Pretty.pretty_com is_printed scope loc v k
  ) output (fun _scope -> []) s
                 )  in
  let ch_out = stdout in
  PPrint.ToChannel.pretty 1. 60 ch_out (Pp.piprogram output);
  flush ch_out

let _ =
  try
    if Array.length Sys.argv > 1 then begin
      if Sys.argv.(1) = "--refactor"
      then begin
        assert (Array.length Sys.argv = 4);
        let instructions = load_refactoring_spec Sys.argv.(2) in
        let lifting_instrs = gen_refactoring_ast instructions Sys.argv.(3) in
        process_flag "input";
        let state = initial_env in
        let state = load_ml_file Sys.argv.(3) state in
        process_flag "lifting";
        let (_,_,_,output) = process_ml_file "<refactoring instructions>" lifting_instrs state in
        print_output output
      end
      else
      let files = List.tl (Array.to_list Sys.argv) in
      let (env, genenv, ornenv,output) = List.fold_left
                                           (fun state -> function
                                                      | s when is_flag s ->
                                                         let root = flag_root s in
                                                         process_flag root; state
                                                      | path -> load_ml_file path state) initial_env files in
      print_output output
    end
    else
      Format.eprintf "Usage: %s inputfile.ml inputfile2.ml ... > output-file.ml@." Sys.argv.(0)
  with
    Error.Error (loc, s) -> print_error loc s; exit 1
