(* From an ornament definition, generate the encoding of the ornament.
   Uses deep pattern matching elimination (\ie. in ML) before
   lifting to \gML in order to share the code of the elimination *)

open Ast
module O = Ast.Orn
open Util
module X = Ast.Explicit
module E = Eml.ExplicitGml
module I = Orn.Implicit
module S = Ast.Scope

let ty = TyArrow(TyVar (id "untyped"), TyVar (id "untyped"))
let sc = Scheme([],ty)

(* register this as an hole ! *)
let absurd_expr () =
  let x = id "_absurd_" in
  IdHashtbl.add Pretty.holemap  x "<absurd>";
  X.Var(x,[]), ty, dummy_loc

let rec pat_to_expr = function
  | (X.PVar(x), ty, loc) -> X.Var(x,[]),ty,loc
  | (X.PCon(c,ts,ps), ty, loc) -> X.Con(c,ts, List.map pat_to_expr ps), ty, loc
  | _ -> failwith "Not a pattern and an expression"

(* Idea: somewhat flatten the sum-of-product for a more intuitive representation.
   Representation:
     sum([t]) -> t
     sum([unit, unit]) -> bool
     sum([unit,t])/sum([t,unit]) -> t option (?)
   otherwise, a either-telescope

  also, remove all units and impossible cases (if any ?)

  -- we should probably use polymorphic variants instead
 *)
type ext =
  | None
  | Unit
  | Thing

let gen_code loc env def =
  (* Substitute on the left ! Compute the skeleton of the left side !
     Substitute on the right ! Keep the same type !
     Generate pattern matchings ! Simplify them !
     Substitute to get the functions !

     We are working with quasi-untyped code.
   *)
  let input = id "input" in
  let cvars = List.map (fun _ -> id "clause") def.O.orn_clauses in
  let ret_ty = TyVar(id "ret_ty") in
  (* Accidentally quadratic *)
  let rec prescan = function
    | X.PZero, _, _ -> None
    | X.PWild, _, _ -> Thing
    | X.PVar _, _ , _ -> Unit
    | X.PAlt(p1,p2), _, _ ->
       begin match prescan p1, prescan p2 with
       | None, t | t, None -> t
       | _ -> Thing
       end
    | X.PCon(_,_,ps), _, _ ->
       let cs = List.map prescan ps in
       if List.mem None cs then None
       else if List.for_all (fun x -> x = Unit) cs then Unit
       else Thing
  in
  (* Invariant: unit extensions do not look at the variable *)
  let rec extension (y,y_ty) = function
    | X.PWild, ty, loc -> X.Var(y,[]), ty, loc
    | X.PVar v, ty, loc -> X.Var(v,[]), ty, loc
    | X.PCon(c, ts, ps), ty, loc -> product_extension y ps ty loc (fun ls -> X.Con(c, ts, ls), ty, loc)
    | X.PZero, ty, loc -> absurd_expr ()
    | X.PAlt(p1, p2), ty, loc -> (* Might be nicer if we create n-way alternatives and n-way products *)
       begin match prescan p1, prescan p2 with
             | t, None -> extension (y, ty) p1
             | None, t -> extension (y, ty) p2
             | _ ->
                let y1, y2 = id "yl", id "yr" in
                X.Match( (X.Var(y,[]), ty, loc),
                         [(X.PCon(sum_l, [], [X.PVar(y1),ty,loc]),ty,loc), extension (y1, ty) p1;
                          (X.PCon(sum_r, [], [X.PVar(y2),ty,loc]),ty,loc), extension (y2, ty) p2] ), ty, loc
       end
  and product_extension y ps ty loc cont =
    let cs = List.map prescan ps in
    let ys = List.mapi (fun i (_,ty,_) -> id ("y"^string_of_int i), ty) ps in
    let u = cont (List.map2 extension ys ps) in
    let ys_truncated, _ = List.split (List.filter (fun (_,x) -> x <> Unit) (List.combine ys cs)) in
    if ys_truncated = [] then u
    else X._LetTuple(ys_truncated,y, u), ty, loc (* let (ys,) = y in u *)
  in
  let ext_var = id "ext_var" in
  let mkp p = (p, ty, dummy_loc) in
  let left_side =
    X.Match((X.Var(input, []), def.O.orn_from_skel, dummy_loc),
            List.map2 (fun (p,q) v -> (p, extension (ext_var,ty) q)) def.O.orn_clauses cvars
              @ (if def.X.orn_fallback then [(mkp X.PWild, absurd_expr())] else [])
    ), ret_ty, dummy_loc in
  let right_side =
      X.Match((X.Var(input, []), def.O.orn_to, dummy_loc),
              List.map2 (fun (p,q) v -> match q with
                                        | X.PZero,_,_ -> (q, absurd_expr ())
                                        | q -> (q, pat_to_expr p)) def.O.orn_clauses cvars), ret_ty, dummy_loc in
  debug "Generated an ornament:\nLeft:\n%a\n\nRight:\n%a\n\n" Pp.term left_side Pp.term right_side;
  let nice_encoder = left_side in
  let left_side =
    Desugaring.(
      try
        if def.X.orn_fallback then untyped_all env left_side
        else untyped_all_ornament env left_side
    with
    | PatternSimplification.OverlappingPatterns -> Error.error loc "Overlapping patterns on left-hand side of ornament definition"
    | PatternSimplification.IncompletePatterns -> Error.error loc "Incomplete pattern matching on left-hand side of ornament definition")
  in
  let right_side = Desugaring.(
    try
      untyped_all_ornament env right_side
    with
    | PatternSimplification.OverlappingPatterns -> Error.error loc "Overlapping patterns on right-hand side of ornament definition"
    | PatternSimplification.IncompletePatterns -> Error.error loc "Incomplete pattern matching on right-hand side of ornament definition")
  in
  (* Let us pretty-print the code before... *)
  say "Generated ornament %a with encoder:\n%a\n\n" Pp.id def.O.orn_name Pp.term nice_encoder;
  let open Eml_compute.ExplicitToGml in
  let orn_loc = dummy_loc in
  let dummy_ty = E.TVar(id "##UNTYPED##") in
  let left_side = E.GAbs(input, (E.GAbs(ext_var, ml_to_gml left_side), dummy_ty, orn_loc)), dummy_ty, orn_loc in
  let right_side = E.GAbs(input, ml_to_gml right_side), dummy_ty, orn_loc in
  let ext_fun = E.TAbs(input, E.TVar(id "__input"), E.TVar(id "__ext")) in
  (ext_fun, left_side, right_side)

let check_left_pat lpat =
  let rec check (p,loc) =
    match p with
    | PVar _ -> ()
    | PWild -> Error.error loc "Left pattern is not an expression (a wildcards appears). Please replace it by a variable."
    | PAlt _ -> Error.error loc "Left pattern is not an expression (an alternative appears)"
    | PZero -> Error.error loc "Left pattern is not an expression (zero appears)"
    | PTuple ps -> List.iter check ps
    | PCon(_, ps) -> List.iter check ps
    | PTyped(p, t) -> check p (* XXX maybe this only makes sense at the first level *)
  in
  match lpat with
  | PCon(_), _ | PTuple _, _ -> check lpat
  | _, loc -> Error.error loc "Left patterns must start with a constructor"

(* Take an ornament given in type-syntax, give its left projection *)
let oty_proj_left oenv loc =
  let rec proj = function
    | TyArrow(t1,t2) -> TyArrow(proj t1, proj t2)
    | TyVar(v) -> TyVar(v)
    | TyProduct(ts) -> TyProduct(List.map proj ts)
    | TyCon(c,ts) ->
       let odef,_,_,_ = List.assoc c oenv.X.env_ornaments in
       if List.length odef.X.orn_args <> List.length ts
       then Error.error loc "Wrong arity for %a." Pp.id c;
       let subst = List.map2 (fun x t -> (x,proj t)) odef.X.orn_args ts in
       let _, subst_ty = X.SubstRename.(make_subst id_term (fun v -> List.assoc v subst)) in
       subst_ty odef.X.orn_from
  in
  proj

(* Take an ornament given in type-syntax, give its right projection *)
let oty_proj_right oenv loc =
    let rec proj = function
    | TyArrow(t1,t2) -> TyArrow(proj t1, proj t2)
    | TyVar(v) -> TyVar(v)
    | TyProduct(ts) -> TyProduct(List.map proj ts)
    | TyCon(c,ts) ->
       let odef,_,_,_ = List.assoc c oenv.X.env_ornaments in
       if List.length odef.X.orn_args <> List.length ts
       then Error.error loc "Wrong arity for %a." Pp.id c;
       let subst = List.map2 (fun x t -> (x,proj t)) odef.X.orn_args ts in
       let _, subst_ty = X.SubstRename.(make_subst id_term (fun v -> List.assoc v subst)) in
       subst_ty odef.X.orn_to
  in
  proj

(* If a mapping is rigid, variables can only be mapped to variables *)
let invert_rigid mapping =
  let inv_map = List.map (function
    | x, TyVar y -> (y, x)
    | x, t -> Error.user_error "Type error: cannot unify type %a with type argument %a." Pp.ty_nice t Pp.id_nice x
  ) mapping in
  (* Check if the mapping is unique *)
  List.iter (fun (y,_) ->
    match List.filter (fun (y',_) -> y = y') inv_map with
    | (_,x)::(_,x')::_ -> Error.user_error "Type error: cannot unify type arguments %a and %a." Pp.id_nice x Pp.id_nice x'
    | _ -> ()
  ) inv_map;
  inv_map

let rec subst_in_pat substs pat =
  let rec su = function
    | X.PVar v, ty, loc ->
       begin try List.assoc v substs with Not_found -> X.PVar v, ty, loc end
    | X.PCon(c, ts, ps), ty, loc ->
       X.PCon(c, ts, List.map su ps), ty, loc
    | _ -> assert false
  in
  su pat

(** Typechecking an ornament definition **)
let typecheck_orndef loc env genv oenv o =
  (* The definition has already been introduced in the stub env oenv *)
  let orn_name = S.(resolve oenv.X.orn_scope NsType o.orn_name) in
  (* Introduce the names in the definition *)
  let orn_args, odef_scope = S.(intros oenv.X.orn_scope NsTyVar (fun x -> x) (fun _ x -> x) o.I.orn_args) in
  let orn_from = Ml_inference.resolve_ty odef_scope o.I.orn_from in
  let orn_to = Ml_inference.resolve_ty odef_scope o.I.orn_to in
  let i_skel = match orn_from with
    | TyCon(i, ts) -> List.assoc i genv.X.skel_map
    | _ -> Error.user_error "Not an ornament of a datatype"
  in
  let skel_descr = List.assoc i_skel genv.X.skel_dataenv in
  (* Examine the clauses *)
  let open Ml_inference in
  let open Ml_inference.Solver in
  (* Fix the tuple/constructor syntax ambiguity *)
  let clauses = List.map (fun (lpat, otypes, rpat) -> (pfix_contuple env lpat, otypes, pfix_contuple env rpat)) o.orn_clauses in
  (* Remove the fallback clause, as we are not able to process it *)
  let orn_fallback, clauses =
    match List.rev clauses with
    | ((PWild,_),[],(PZero,_))::cs -> true, List.rev cs
    | _ -> false, clauses
  in
  (* Check:
     - pattern on the left has the right shape
     - pattern on the left and right have the same variables
  *)
  List.iter (fun (lpat,otypes,rpat) ->
    let (_,loc) = lpat in
    check_left_pat lpat;
    begin match rpat with
    | PZero, _ -> ()
    | _ , loc ->
       (* Rule: no duplicate variables, except for variables starting with an underscore *)
       if
         List.exists (fun x -> true) (list_diff (patvars lpat) (patvars rpat))
       then Error.error loc "Variables do no match on left and right of pattern."
    end;
    let lvars = patvars lpat in
    List.iter (fun (x,_) -> if not (List.mem x lvars)
      then Error.error loc "Variable %a is given a type but not bound by this pattern." Pp.user_id x
    ) otypes
  ) clauses;
  (* Transform the clauses from a 3-element representation (lpat, otypes, rpat)
     to (lpat, otypes, re-match set, rpat), it's much easier to process afterwards *)
  let clauses = List.map (fun (lpat, otypes, rpat) ->
    let otypes = List.map (fun (var, oty) -> (var, resolve_ty odef_scope oty)) otypes in
    let rec proc = function
      (* Typed things become variables that are typed & re-matched *)
      | PTyped(p, t), loc ->
         let var = id "z" in
         (PVar (FromId var), loc), [(var, p, Ml_inference.resolve_ty odef_scope t)]
      (* Other things just recurse and accumulate *)
      | PCon(c, ps), loc ->
         let (ps, subpats) = List.split (List.map proc ps) in
         (PCon(c,ps), loc), List.concat subpats
      | PTuple(ps), loc ->
         let (ps, subpats) = List.split (List.map proc ps) in
         (PTuple ps, loc), List.concat subpats
      | PVar x, loc -> (PVar x, loc), []
      | _ -> assert false (* unexpected construct *)
    in
    let lpat, things = proc lpat in
    lpat,
    otypes @ List.map (fun (v, _, t) -> (FromId v,t)) things,
    List.map (fun (v, p, _) -> (v,p)) things,
    rpat
  ) clauses in
  (* Determine the base type of each variable by inference.
     This constraint is solved in a separate pass to gather the types.
     The types are then:
      - either compared with the ornament given for this variable
      - or used as identity ornaments
  *)
  let cstr_left_from =
    freshen orn_args (fun mapping ->
      exist_ (fun from_v ->
        ty_encode mapping from_v orn_from ^^
          conj_list
          (List.map
             (fun (lpat, otypes, midpat, rpat) ->
               let xs, pat_scope = Scope.(intros env.scope NsVar (fun x -> x) (fun x x' -> (x,x')) (patvars lpat)) in
               let rec abstract k = function
                 | [] -> k ^^ pure []
                 | (x,x')::xs -> exist (fun v -> def x' v (abstract k xs)) >>= fun (t,bs) -> ((x,t)::bs)
               in
               abstract ((inference_things env)#pattype pat_scope lpat from_v) xs >>=
                 (fun ltypes -> (lpat, ltypes, otypes, midpat, rpat))
             ) clauses)
      )
    ) >>= fun (vtys, clauses) ->
    let left_unmap = invert_rigid (List.combine orn_args vtys) in
    let _, unsubst_ty = X.SubstRename.(make_subst id_term (fun v -> TyVar(List.assoc v left_unmap))) in
    List.map
      (fun (lpat,ltypes,otypes,midpat,rpat) ->
        (lpat, List.map (fun (x,t) -> (x, unsubst_ty t)) ltypes, otypes, midpat, rpat)
      ) clauses
  in
  let _, clauses =
    try solve false (let0 cstr_left_from)
    with Unify(t1,t2) -> Error.error loc "Type error in left-hand side pattern: cannot unify %a and %a" Pp.ty_nice t1 Pp.ty_nice t2 in
  (* Resolve otypes, check coherence with left types or fill missing with identity
    -> we obtain an ornament type for all variables crossing the base/orn boundary *)
  let clauses = List.map (fun (lpat, ltypes, otypes, midpat, rpat) ->
    let _, loc = lpat in
    let otypes =
      List.map (fun (x,lty) ->
        match List.assoc x otypes with
        | oty ->
           let lty_o = oty_proj_left oenv loc oty in
           if not (lty_o = lty) then
             Error.error loc "Base type of %a is %a, but %a expected" Pp.ty_nice oty Pp.ty_nice lty_o Pp.ty_nice lty;
           (x, oty)
        | exception Not_found -> (x, lty) (* Identity ornament *)
      ) ltypes
    in
    (lpat, otypes, midpat, rpat)
  ) clauses in
  (* Typecheck left/middle/right in the new environment so that we get a well-typed skeleton *)
  let cstr_to =
    freshen orn_args (fun mapping ->
      exist (fun skel_v ->
        freshen skel_descr.X.skel_args (fun l -> skel_v --- S.TyCon(i_skel, List.map snd l)) ^^
        exist_ (fun to_v ->
          ty_encode mapping to_v orn_to ^^
            conj_list
            (List.map
               (fun (lpat, otypes, midpat, rpat) ->
                 let loc = snd lpat in
                 let lpat_skel = match lpat with
                   | PCon(c,ps), loc ->
                      let c = Scope.(resolve env.scope NsCons c) in
                      let c_skel = List.assoc c genv.X.skel_map in
                      PCon(FromId c_skel, ps), loc
                   | PTuple(ps), loc ->
                      let c = tuple_con.(List.length ps) in
                      let c_skel = List.assoc c genv.X.skel_map in
                      PCon(FromId c_skel, ps), loc
                   | _ -> assert false (* Excluded by check_left_pat *)
                 in
                 let otypes, pat_scope = Scope.(intros env.scope NsVar (fun (x,oty) -> x) (fun (_,oty) x' -> (x',oty)) otypes) in
                 let right_types = List.map (fun (x,oty) -> (x, oty_proj_right oenv loc oty)) otypes in
                 let rec abstract k = function
                   | [] -> k
                   | (x,oty)::xs ->
                      exist_ (fun v -> ty_encode mapping v oty ^^ def x v (abstract k xs))
                 in
                 let rec abstract_ k = function
                   | [] -> k
                   | x::xs -> exist_ (fun v -> def x v (abstract_ k xs))
                 in
                 (* These are userid but we still need to scope them I think *)
                 let midvars = List.concat (List.map (fun (_, p) -> patvars p) midpat) in
                 let midvars, mid_scope = Scope.(intros pat_scope NsVar (fun x -> x) (fun _ x' -> x') midvars) in
                 abstract ((inference_things env)#pattype pat_scope lpat_skel skel_v) otypes ^^
                   abstract (
                     (* Here we still need to match the middle patterns! *)
                     exist_ (fun skel_proj_v-> (inference_things env)#pattype pat_scope lpat_skel skel_proj_v) ^&
                       abstract_ (
                         conj_list (List.map (fun (v, midpat) ->
                           exist_ (fun tau ->
                             instance v tau ^^
                               pure v ^& (inference_things env)#pattype mid_scope midpat tau
                         )) midpat) ^&
                         (inference_things env)#pattype mid_scope rpat to_v
                       ) midvars
                   ) right_types
                                                                                  >>=
                 fun (lpat, (midpats, rpat)) ->
                 (subst_in_pat midpats lpat, rpat)
               ) clauses
            )
        )
      )
    ) >>= fun (vtys, (skel_from, clauses)) ->
    let right_unmap = invert_rigid (List.combine orn_args vtys) in
    (* XXX we are breaking typing with the escape clause here *)
    let _, unsubst_ty, unsubst_pa = X.SubstRename.(make_subst_full id_term (fun v ->
      try TyVar(List.assoc v right_unmap)
      with Not_found -> TyVar(v))
    ) in
    unsubst_ty skel_from,
    List.map (fun (lpat_skel, rpat) -> unsubst_pa lpat_skel, unsubst_pa rpat) clauses
  in
  let _, (orn_from_skel, orn_clauses) =
    try
      solve false (let0 cstr_to)
    with
    | Unify(t1,t2) -> Error.error loc "Type error in ornamentation clause: cannot unify %a and %a" Pp.ty_nice t1 Pp.ty_nice t2
  in
  say "[type ornament _ %a : %a => %a] is defined@." Pp.id_nice orn_name Pp.ty_nice orn_from_skel Pp.ty_nice orn_to;
  X.{orn_name; orn_args;orn_from;orn_from_skel;orn_to;orn_clauses;orn_fallback}

(** Elaborate a group of ornament definitions **
  - Introduce the ornament names into the scope.
  - Add "stub" definitions into the environment.
  - Elaborate each definition individually in the stub environment.
**)
let elaborate_orndef loc env genv oenv os =
  (* Generate a typed version of the `os': code, and skeleton arguments *)
  (* Introduce ornament names without the code at first *)
  let _, odef_scope = S.(intros oenv.Orn.orn_scope NsType (fun o -> o.I.orn_name) (fun _ _ -> ())) os in
  let orns_stub = List.map (fun o ->
    let orn_name = S.(resolve odef_scope NsType o.orn_name) in
    let orn_args, odef_scope = S.(intros odef_scope NsTyVar (fun x -> x) (fun _ x -> x) o.I.orn_args) in
    let orn_from = Ml_inference.resolve_ty odef_scope o.I.orn_from in
    let orn_to = Ml_inference.resolve_ty odef_scope o.I.orn_to in
    let orn_from_skel = TyVar(id "dummy") in
    let orn_clauses = [] in
    let orn_fallback = false in
    let dummy_ty = E.TVar(id "dummy") in
    let dummy_expr = E.Var(id "dummy"), dummy_ty, dummy_loc in
    orn_name,
    (X.{ orn_name; orn_args; orn_from; orn_from_skel; orn_to; orn_clauses; orn_fallback },
     dummy_ty, dummy_expr, dummy_expr)
  ) os in
  let oenv_stub = Orn.{oenv with orn_scope = odef_scope; env_ornaments = orns_stub@oenv.env_ornaments} in
  let os = List.map (typecheck_orndef loc env genv oenv_stub) os in
  (* Typecheck in the stub environment, generate code and add to the normal environment *)
  let orns = List.map (fun o -> let (ext,left,right) = gen_code loc env o in
                                o.X.orn_name, (o, ext, left, right) ) os in
  Orn.{oenv with orn_scope = odef_scope; env_ornaments = orns@oenv.env_ornaments}

let orndef_identity env generic_env i =
  (* I only use generic_env *)
  let i_skel = List.assoc i generic_env.X.skel_map in
  let skel_desc = List.assoc i_skel generic_env.X.skel_dataenv in
  let orn_args = skel_desc.X.skel_id_args in
  let ty_args = List.map (fun x -> TyVar(x)) orn_args in
  let orn_ty = TyCon(i, ty_args) in
  let skel_args = skel_desc.X.skel_id_inst in
  let orn_ty_skel = TyCon(i_skel, skel_args) in
  let clauses =
    List.map
      (fun (c_skel, args_ty) ->
       let c = List.assoc c_skel generic_env.X.skel_unmap in
       let vars = List.map (fun ty -> (X.PVar(id "x"), TyVar (id "dummy"), dummy_loc)) args_ty in
       (X.PCon(c_skel, skel_args, vars), orn_ty_skel, dummy_loc),
       (X.PCon(c, ty_args, vars), orn_ty, dummy_loc)
      ) skel_desc.X.skel_conses
  in
  X.{
    orn_name = i;
    orn_args;
    orn_from = orn_ty;
    orn_from_skel = orn_ty_skel;
    orn_to = orn_ty;
    orn_clauses = clauses;
    orn_fallback = false
  }

let elaborate_identity loc env genv oenv is =
  let orn_scope =
    List.fold_left (fun orn_scope i -> Scope.(intro_named orn_scope NsType (UserId i.id_name) i))
      oenv.X.orn_scope is in
  let orns = List.map (fun i ->
    let idef = orndef_identity env genv i in
    let (ext,left,right) = gen_code loc env idef in
    (idef.X.orn_name, (idef, ext, left, right))
  ) is in
  Orn.{oenv with env_ornaments = orns@oenv.env_ornaments; orn_scope}

(* Ideally we would want to type the patch but this is simpler *)
let name_patch scope vars e =
  (* Filter out non-user variables. They should not be used in patches *)
  debug "Vars where %a@." Pp.(seq "," id) vars;
  let vars = List.filter (fun x -> x.id_user || (String.length x.id_name > 0 && x.id_name.[0] = '_')) vars in
  debug "Filtered to %a@." Pp.(seq "," id) vars;
  let open Scope in
  let scope = {scope with vars = List.map (fun x -> (UserId(x.id_name), x)) vars @ scope.vars} in
  let ty = TyVar(id "") in
  let sc = Scheme([], ty) in
  let rec cv_aux scope = function
    | Var(x) -> X.Var(resolve scope NsVar x, [])
    | Let(is_meta,isrec, bs, e) ->
       let vars, in_scope = intros scope NsVar (fun (x,_) -> x) (fun _ x -> x) bs in
       let let_scope = if isrec then in_scope else scope in
       X.Let(is_meta, isrec, [], List.map2 (fun x (_,e) -> (x, sc, cv let_scope e)) vars bs, cv in_scope e)
    | Match(e, cs) ->
       X.Match(cv scope e, List.map (clause scope) cs)
    | Con(c,es) -> X.Con(resolve scope NsCons c,[], List.map (cv scope) es)
    | Abs(x, e) ->
       let x, scope = intro scope NsVar x in
       X.Abs(x, ty, cv scope e)
    | App(e, f) -> X.App(cv scope e, cv scope f)
    | Tuple(es) -> X._Tuple(List.map (cv scope) es)
    | Label(l, e) -> X.Mark(X.NiceName l, cv scope e)
  and cv scope (e,loc) = cv_aux scope e, ty, loc
  and clause scope (p, e) =
    let vars = Ml_inference.patvars p in
    let _, scope = intros scope NsVar (fun x -> x) (fun _ x -> x) vars in
    (pat scope p, cv scope e)
  and pat scope(p, loc) = pat_aux scope p, ty, loc
  and pat_aux scope = function
    | PZero -> X.PZero
    | PVar(x) -> X.PVar(resolve scope NsVar x)
    | PWild -> X.PWild
    | PAlt(p,q) -> X.PAlt(pat scope p, pat scope q)
    | PTuple(ps) -> X._PTuple(List.map (pat scope) ps)
    | PCon(c,ps) -> X.PCon(resolve scope NsCons c, [], List.map (pat scope) ps)
    | PTyped(p, t) -> let p, _, _ = pat scope p in p
  in
  cv scope e
